/*==============================================================================
// 
//	File: ANECB.h
//
//	Argus Numerical Environments - Plug In Extension, Release 1.0
// 
//	Copyright  1996-97 Argus Holdings Ltd.  All rights reserved.
//
//		THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF Argus Holdings Ltd..
//		The copyright notice above does not evidence any
//		actual or intended publication of such source code
//
//==============================================================================*/

#ifndef _ANECB_H_
#define _ANECB_H_

#ifndef _ANEPIE_H_
#include "ANEPIE.h"
#endif /*_ANEPIE_H_*/

#ifdef __cplusplus
extern "C"
{
#endif

/* Basic calls */
/*==========*/

void ANE_GetVersion(ANE_PTR aneHandle, int* major, int* minor, int* update, char* versionBuf, int bufLen);
	/* get version information for the ArgusNE application */

void ANE_GetPIEProjectHandle(ANE_PTR aneHandle, ANE_PTR* pieHandle);
	/* get a pointer to the private information of the current document's Project */

void ANE_ProcessEvents(ANE_PTR aneHandle);
	/* give a chance to ArgusNE and the rest of the system to handle user events */

void ANE_MakeDirty(ANE_PTR aneHandle);
	/* mark that the document was touched (turn on Save menu etc. ) */


void ANE_ImportTextToLayer(ANE_PTR handle, ANE_STR textToImport);
	/* calls the current layer to import the text in 'textToImport' as an import file */

void ANE_ExportTextFromOtherLayer(ANE_PTR handle, ANE_PTR layerHandle, ANE_STR* textExported);
	/* calls the layer pointed by 'layerHandle' to export it's information into 'textExported' pointer */

/* Memory allocation functions */

ANE_PTR ANE_MemAlloc(ANE_PTR handle, ANE_INT32 size);
	/* allocate size bytes of memory */

void ANE_MemDelete(ANE_PTR handle, ANE_PTR ptr);
	/* delete chunk of memory */


/* Protection functions */
/*================*/

void ANE_ProtGetRegCode(ANE_PTR handle, ANE_STR regCodeBuf, ANE_INT32 bufLen);
	/* return into regCodeBuf min(regcode length, bufLen-1) characters of the RegCode */

void ANE_ProtGetUniqueID(ANE_PTR handle, ANE_STR uniqueIDBuf, ANE_INT32 bufLen);
	/* return into uniqueIDBuf min(unique id  length, bufLen-1) characters of the Worksstation Unique ID */

/* Directory and file functions */
ANE_BOOL ANE_DirectoryGet(ANE_PTR handle, ANE_STR caption, ANE_STR path, ANE_STR dirBuf, ANE_INT32 bufLen);
	/* pop a dialog to the user, allowing selection of a directory. return true if directory was 
		selected, false if directory was not selected. If directory was selected, a null terminated string is returned in buf */
		
ANE_BOOL ANE_FileGet(ANE_PTR handle, ANE_STR caption, ANE_STR path, ANE_STR defName, ANE_BOOL forExport, ANE_STR dirBuf, ANE_INT32 bufLen);
	/* pop a dialog to the user, allowing selection of a file. return true if a file was 
		selected, false if no file was selected. If file was selected, a null terminated string is returned in buf */

ANE_BOOL ANE_FileGetEx(ANE_PTR handle, ANE_STR caption, ANE_STR path, ANE_STR defName, ANE_BOOL forExport, ANE_STR dirBuf, ANE_INT32 bufLen, ANE_STR extensionMask, ANE_STR extensionCaption, ANE_INT32 macType);
	/* pop a dialog to the user, allowing selection of a file. return true if a file was 
		selected, false if no file was selected. If file was selected, a null terminated string is returned in buf ,
		The default extenstion of the file is defined by extensionMask, and described by extensionCaption. 
		macType is used to select the files on the macintosh, where extensions don't apply */
		
void ANE_DirectorySetCurrent(ANE_PTR handle, ANE_STR currentDirectory);
	/* set the current directory */
	
void ANE_DirectoryGetCurrent(ANE_PTR handle, ANE_STR buf, ANE_INT32 bufLen);
	/* return the name of the directory */
	

/* Layer Handling callbacks */
/*===================*/

ANE_PTR ANE_LayerGetHandleByName(ANE_PTR handle, ANE_STR name);
	/* return a handle for the layer with the name 'name', 0 if no layer with that name */
	
ANE_PTR ANE_LayerAddByTemplate(ANE_PTR handle, ANE_STR layerTemplate, ANE_PTR afterLayerHandle);
	/* parse a Layer Definition Language template, and adds a layer.
			New layers will be added after 'afterLayerHandle', if 'afterLayerHandle' is zero,
			new layers will be added at the end of the layers list.
			If a layers were added, return the handle to the first added layer.
			If parsing failed, returns 0 */

void ANE_LayerAddParametersByTemplate(ANE_PTR handle, ANE_PTR atLayerHandle, ANE_STR paramsTemplate, ANE_INT32 afterParameter);
	/* parse a Layer Definition Language template containing only parameters, and addthem to 'atLayerHandle'.
			New parameters will be added after 'afterParameter', if 'afterParameter' is negative,
			new parameters will be added at the end of the parameters list. */

ANE_BOOL ANE_LayerClear(ANE_PTR handle,  ANE_PTR layerHandle, ANE_BOOL selectedOnly);
	/* clear all objects in the specified layer. Returns true if there were objects in
		the layer, return false if no objects were cleared.*/

ANE_BOOL ANE_LayerRemove(ANE_PTR handle, ANE_PTR layerHandle, ANE_BOOL forceRemoval);
	/* remove the specified layer. If forceRemoval is TRUE, remove the layer even 
		if it is a geographic layer that other layer is using. Return TRUE if a layer was removed*/

ANE_BOOL ANE_LayerRename(ANE_PTR aneHandle, ANE_PTR layerHandle, ANE_STR newName);
	/* if the layer is renamed, return TRUE*/

ANE_INT32 ANE_LayerGetNumParameters(ANE_PTR aneHandle, ANE_PTR layerHandle, enum EPIEParameterType parameterType);
	/* return the number of parameter of the layer*/

ANE_INT32 ANE_LayerGetParameterByName(ANE_PTR aneHandle, ANE_PTR layerHandle, ANE_STR parameterName);
	/* return the index of the parameter with the given  layer, return -1*/

ANE_BOOL ANE_LayerRemoveParameter(ANE_PTR aneHandle, ANE_PTR layerHandle, ANE_INT32 parameterIndex, ANE_BOOL forceRemoval);
	/* remove the specified parameter. If forceRemoval is TRUE, remove the parameter even 
		if is is used by other layers. Return TRUE if a parameter was removed*/

ANE_BOOL ANE_LayerRenameParameter(ANE_PTR aneHandle, ANE_PTR layerHandle, ANE_INT32 parameterIndex, ANE_STR newParameterName);
	/* if the parameter was renamed, return TRUE*/

void ANE_LayerSetParameterExpression(ANE_PTR aneHandle, ANE_PTR layerHandle, ANE_INT32 parameterIndex, ANE_STR newParameterExpression);
	/* set the parameter's expression*/

void ANE_ImportTextToLayerByHandle(ANE_PTR handle, ANE_PTR layerHandle, ANE_STR textToImport);
	/* calls a layer to import the text in 'textToImport' as an import file */

ANE_STR* ANE_LayerGetNamesByType(ANE_PTR handle, enum EPIELayerType requestedTypes);
	/* get the names of all layers of a given type. The memory pointer returned by this
		functio must be released using ANE_MemDelete. */




/* Data Layer Callbacks */
/*===================*/

void ANE_DataLayerSetData(ANE_PTR 		handle, 
						ANE_PTR 		dataLayerHandle, 
						ANE_INT32		numPoints,
						ANE_DOUBLE*		posX,
						ANE_DOUBLE*		posY,
						ANE_INT32		numDataParameters,
						ANE_DOUBLE**	dataParameters,
						ANE_STR*			paramNames);

	/* Set scattered data in a data layer.
		There are numPoints points of data.
		Point i is at location posX[i] posy[i].
		
		Each point has numDataParameters data items attached to it.
		the first parameter of point i is dataParameters[p][i]
	*/
						
void ANE_DataLayerSetTriangulatedData(ANE_PTR 		handle, 
						ANE_PTR 		dataLayerHandle, 
						ANE_INT32		numPoints,
						ANE_DOUBLE*		posX,
						ANE_DOUBLE*		posY,
						ANE_INT32 		numTriangles,
						ANE_INT32*		node0,
						ANE_INT32*		node1,
						ANE_INT32*		node2,
						ANE_INT32		numDataParameters,
						ANE_DOUBLE**	dataParameters,
						ANE_STR*			paramNames);
	
	/* Set triangulated data in a data layer.
		There are numPoints points of data.
		Point i is at location posX[i] posy[i].
		
		Each point has numDataParameters data items attached to it.
		The first parameter of point i is dataParameters[p][i]
		
		The triangulation information is defined by node0, node1, node2.
		There are numTriangles triangle.
		Triangle j has three vertices: node0[j] node1[j] and node2[j], each of the is
		an index to a point, that is between 0 and numPoints-1.

		if paramNames is NULL, parameters get the namee "Imported Parameter 1" etc.
		if paramNames is not NULL, parameters get the names paramNames[i]
	*/


/* User Interface Callbacks */
/*===================*/

void ANE_UserMessageOk(ANE_PTR handle, ANE_STR message);
	/* displays a dialog box with the message and an OK button */
	
ANE_BOOL ANE_UserMessageOkCancel(ANE_PTR handle, ANE_STR prompt);
	/* displays a dialog box with the prompt, and two buttons: Ok and Cancel.
		returns true if the user selected Ok and false if the user selected cancel */
		
ANE_INT32 ANE_UserMesageYesNoCancel(ANE_PTR handle, ANE_STR prompt);
	/* displays a dialog box with the prompt and three buttons: Yes, No and Cancel.
		returns 1 for Yes
				2 for No
				3 for Cancel
	*/

ANE_INT32 ANE_UserSelectItem(ANE_PTR handle, ANE_STR* items, ANE_STR title, ANE_STR prompt);
	/* displays a dialog box with a list of items. items is an array of pointers to char, 
	terminated by a null pointer.
	The message window will have title as its title, and prompt as the prompt message to the user
	return index of item (0 for first, 1 for second, etc.), or -1 if no item was selected.*/

ANE_BOOL ANE_UserSelectMultipleItems(ANE_PTR handle, ANE_STR* items, ANE_STR title, ANE_STR prompt, ANE_INT32* returnNumSelected, ANE_INT32** returnSelectedItems);
	/* displays a dialog box with a list of items. items is an array of pointers to char, 
	terminated by a null pointer.
	The message window will have title as its title, and prompt as the prompt message to the user
	
	the function returns true if the user pressed the Ok button
	the function returns false if the user pressed the Cancel button, or closed the window
	
	if the function returns true, the integer pointed by returnNumSelected is set to the number of selected items
	in addition, the pointer to integet pointed by returnSelectedItems is set to a pointer to array of indices of the selected items.
	it is the caller function responsibility to free the memory allocated, by calling ANE_MemDelete(aneHande, *returnSelectedItems)
	
	*/


void ANE_EvaluateStringAtLayer(ANE_PTR handle, ANE_PTR layerHandle, enum EPIENumberType type, ANE_STR string, ANE_PTR result);
	/* evaluates 'string' to yield value of type 'type', use the evaluation centext 
	of layer pointed by 'layerHandle' the result will be stored into 'result' */


#ifdef __cplusplus
}
#endif



#endif /* _ANECB_H_ */

/*==============================================================================
//	End File: ANECB.h
//==============================================================================*/
