/******************************************************************************\
* 
* This project was contributed by  Dr. Clifford Voss. of the USGS.
*
*  MODULE:      SUTRA_PIE.C
*
*  PURPOSE:     To provide a Sutra PIE template. 
*
*  First Version: June 1996 by Yaacov Ben-Israel.
*	Modified Sept 1996 - fix some minor interface bugs.
*	Modified Feb 1997 - add front panel "SUTRA MODEL SETUP"
*
*  COMMENTS:   
\******************************************************************************/
#include "stdafx.h" 
#include <windows.h>
#include "sutra_pie.h"
#include <math.h>
#include <winuser.h>
#include <winbase.h>
#include <string.h>
#include <stdio.h>
#include <strstrea.h>

    

#include "sutra.h"
#include "TabbedDialog.h"
#include "Pane1.h"
#include "Pane2.h"
#include "Pane3.h"
#include "Pane4.h"
#include "Pane5.h"
#include "Pane6.h"
#include "Pane7.h"
#include "Pane8.h"
#include "RunSUTRADialog.h"

#include "ModelSetupDialog.h"

#include "ANECB.h"

#include "BaseDialog.h"

#include "SutraPostProc_pie.h"

//==============================================================================
// SUTRA Project accessory functions
//==============================================================================
static void SReplaceString(CString& str, const CString& from, const CString& to)
{
	ASSERT(from != "");

	int index;
	while ( (index = str.Find(from)) >= 0)
	{
		str = str.Left(index) + to + str.Mid(index+from.GetLength());
	}
}

static void SReplaceInt(CString& str, const CString& from, long toInt)
{
	char buf[20];
	sprintf(buf, "%ld", toInt);
	SReplaceString(str, from, buf);
}

static void SReplaceDouble(CString& str, const CString& from, double toDouble)
{
	char buf[30];
	int MININT = -2147483647;
	int MAXINT = 2147483648;
	if ((toDouble > MAXINT)||(toDouble < MININT))
	{
		sprintf(buf, "%.15le", toDouble);
	}
	else
	{
		sprintf(buf, "%.15lg", toDouble);
	};
	

	//sprintf(buf, "%.15lg", toDouble);
	SReplaceString(str, from, buf);
}

static void SReadDefLong(const CString& defaultValues, const char* varName, long& variable)
{
	long index;
	if ( (index = defaultValues.Find(varName)) >= 0)
	{
		int start = index + strlen(varName);
		while (start < defaultValues.GetLength() && (defaultValues[start] == ' ' ||
							defaultValues[start] == '\t'))
					start++;

		if (start < defaultValues.GetLength() && defaultValues[start] == '=')
			start++;

		sscanf((const char*)defaultValues+start, "%ld", &variable);
	}
}

static void SReadDefBool(const CString& defaultValues, const char* varName, BOOL& variable)
{
	long index;
	if ( (index = defaultValues.Find(varName)) >= 0)
	{
		int start = index + strlen(varName);
		while (start < defaultValues.GetLength() && (defaultValues[start] == ' ' ||
							defaultValues[start] == '\t'))
					start++;

		if (start < defaultValues.GetLength() && defaultValues[start] == '=')
			start++;

		sscanf((const char*)defaultValues+start, "%d", &variable);
	}
}

static void SReadDefDouble(const CString& defaultValues, const char* varName, double& variable)
{
	long index;
	if ( (index = defaultValues.Find(varName)) >= 0)
	{
		int start = index + strlen(varName);
		while (start < defaultValues.GetLength() && (defaultValues[start] == ' ' ||
							defaultValues[start] == '\t'))
					start++;

		if (start < defaultValues.GetLength() && defaultValues[start] == '=')
			start++;

		sscanf((const char*)defaultValues+start, "%lf", &variable);
	}
}

static void SReadDefInt(const CString& defaultValues, const char* varName, int& variable)
{
	long index;
	if ( (index = defaultValues.Find(varName)) >= 0)
	{
		int start = index + strlen(varName);
		while (start < defaultValues.GetLength() && (defaultValues[start] == ' ' ||
							defaultValues[start] == '\t'))
					start++;

		if (start < defaultValues.GetLength() && defaultValues[start] == '=')
			start++;

		sscanf((const char*)defaultValues+start, "%d", &variable);
	}
}

static void SReadDefString(const CString& defaultValues, const char* varName, CString& variable)
{
	long index;
	long nameLen = strlen(varName);
	if ( (index = defaultValues.Find(varName)) >= 0)
	{
		int start = index + strlen(varName);
		while (start < defaultValues.GetLength() && (defaultValues[start] == ' ' ||
							defaultValues[start] == '\t'))
					start++;

		if (start < defaultValues.GetLength() && defaultValues[start] == '=')
			start++;

		while (start < defaultValues.GetLength() && (defaultValues[start] == ' ' ||
							defaultValues[start] == '\t'))
					start++;

		long endIndex = defaultValues.Mid(start).Find('\n');
		if (endIndex >= 0)
			variable = defaultValues.Mid(start, endIndex);
	}
}

static BOOL SReadFromFile(const CString& requestedFileName, CString& theString)
{
	char fileName[MAX_PATH];
	HINSTANCE hInstance = AfxGetInstanceHandle( );
	GetModuleFileName(hInstance, fileName, MAX_PATH);

	int i = strlen(fileName) - 1;
	for (; i > 0 && fileName[i] != '\\' && fileName[i] != '/'  ; i--)
		;
	
	ASSERT(i > 0);
	fileName[i+1] = '\0';
	strcat(fileName, requestedFileName);
	
	CStdioFile file;
	
	if (!file.Open(fileName, CFile::modeRead|CFile::typeText))
	{
		return FALSE;
	}

	long len = file.GetLength();
	char* buf = new char[len+1];
	len = file.Read(buf, len);
	buf[len] = '\0';
	theString = CString(buf);
	return TRUE;
}

static CString sDefaultIni = 
"[GENERAL]\n\
@C or T@=C or T\n\
@C or degC@=C or degC\n\
@Energy or Solute@=Energy or Solute\n\
@energy or solute@=energy or solute\n\
@ENERGY_OR_SOLUTE@=ENERGY_OR_SOLUTE\n\
@(M or E)@=(M or E)\n\
@conc_or_temp@=conc_or_temp\n\
@CONC_OR_TEMP@=CONC_OR_TEMP\n\
@Conc or Temp@=Conc or Temp\n"

"@Permeability@=Permeability\n\
@max_min_units@=L^2\n\
@fluid_source_units@=M/s\n\
@solute_source_units@=M/s\n\
@sources_of_e_or_s_units@=(M or E)\n\
@spec_pressure_units@=M/(L s^2)\n\
@Pressure@=Pressure\n\
@pressure@=pressure\n\
@PRESSURE@=PRESSURE\n\
@Pressure boundary condition factor@=Pressure boundary condition factor\n"

"@initial_pressure_units@=M/(L s^2)\n\
@UNSATURATED PROPERTIES LAYER@=Layer\n\
@MESH DENSITY LAYER@=Layer\n\
@FISHNET MESH LAYOUT LAYER@=Layer\n\
@Mesh Density@=Mesh Density\n\
@THICKNESS LAYER@=Layer\n"

"[SOLUTE-VAR]\n\
@C or T@=C\n\
@C or degC@=C\n\
@Energy or Solute@=Solute\n\
@energy or solute@=solute\n\
@ENERGY_OR_SOLUTE@=SOLUTE\n\
@(M or E)@=M\n\
@conc_or_temp@=concentration\n\
@CONC_OR_TEMP@=CONCENTRATION\n\
@Conc or Temp@=Concentration\n"

"@Permeability@=Permeability\n\
@max_min_units@=L^2\n\
@fluid_source_units@=M/s\n\
@sources_of_e_or_s_units@=M\n\
@spec_pressure_units@=M/(L s^2)\n\
@Pressure@=Pressure\n\
@pressure@=pressure\n\
@PRESSURE@=PRESSURE\n\
@Pressure boundary condition factor@=Pressure boundary condition factor\n"


"[SOLUTE-CONST]\n\
@DRWDU@ = 0\n\
@RHOW0@ = 1\n\
@URHOW0@ = 0\n\
@VISCO@ = 1\n\
@GRAVX@ = 0\n\
@GRAVY@ = 0\n\
@CW@ = 1\n"

"@C or T@=C\n\
@C or degC@=C\n\
@Energy or Solute@=Solute\n\
@energy or solute@=solute\n\
@ENERGY_OR_SOLUTE@=SOLUTE\n\
@(M or E)@=M\n\
@conc_or_temp@=concentration\n\
@CONC_OR_TEMP@=CONCENTRATION\n\
@Conc or Temp@=Concentration\n"

"@Permeability@=Hydraulic Conductivity\n\
@max_min_units@=L/s\n\
@fluid_source_units@=L^3/s\n\
@sources_of_e_or_s_units@=(C*L^3)\n\
@spec_pressure_units@=L\n\
@Pressure@=Hydraulic Head\n\
@pressure@=hydraulic_head\n\
@PRESSURE@=HYDRAULIC_HEAD\n\
@Pressure boundary condition factor@=Hydraulic Head\n"

"[ENERGY]\n\
@SIMULA@ = SUTRA ENERGY TRANSPORT\n\
@C or T@=degC\n\
@C or degC@=degC\n\
@Energy or Solute@=Energy\n\
@energy or solute@=energy\n\
@ENERGY_OR_SOLUTE@=ENERGY\n\
@(M or E)@=E\n\
@conc_or_temp@=temperature\n\
@CONC_OR_TEMP@=TEMPERATURE\n\
@Conc or Temp@=Temperature\n"

"@Permeability@=Permeability\n\
@max_min_units@=L^2\n\
@fluid_source_units@=M/s\n\
@sources_of_e_or_s_units@=E\n\
@spec_pressure_units@=M/(L s^2)\n\
@Pressure@=Pressure\n\
@pressure@=pressure\n\
@PRESSURE@=PRESSURE\n\
@Pressure boundary condition factor@=Pressure boundary condition factor\n"


"[USER_SPEC]\n\
@THICK_VALUE@=Thickness\n\
@thickness value@=1\n\
@THICKNESS LAYER@=Layer\n"

"[CYLINDRICAL]\n\
@THICK_VALUE@=2*3.14159265358979*x()\n\
@thickness value@=2*3.14159265358979*x()\n\
@THICKNESS LAYER@=DontIncludeLayer\n"


"[SATURATED]\n\
@UNSATURATED PROPERTIES LAYER@=DontIncludeLayer\n\
@IUNSAT@=0\n"

"[SAT_UNSAT]\n\
@UNSATURATED PROPERTIES LAYER@=Layer\n\
@IUNSAT@=1\n"

"[FISHNET_MESH]\n\
@MESH DENSITY LAYER@=DontIncludeLayer\n\
@FISHNET MESH LAYOUT LAYER@=Layer\n\
@Mesh Density@=Domain Outline\n"

"[IRREGULAR_MESH]\n\
@MESH DENSITY LAYER@=Layer\n\
@FISHNET MESH LAYOUT LAYER@=DontIncludeLayer\n\
@Mesh Density@=Mesh Density\n"


"[AREAL]\n\
@GRAVX@=0\n\
@GRAVY@=0\n\
@THICKNESS LAYER@=Layer\n"

"[CROSS_SECTIONAL]\n"
;
static BOOL
SReadIniSection(const CString& requestedFileName, const CString& requestedSection, CString& theString)
{
	if (!SReadFromFile(requestedFileName, theString))
	{
		theString = sDefaultIni;
	}

	long start = theString.Find("["+requestedSection+"]");
	if (start < 0)
	{		
		theString = "";
		return FALSE;
	}

	theString = theString.Mid(start+requestedSection.GetLength()+2);

	long end = theString.Find('[');
	if (end >= 0)
	{
		theString = theString.Left(end);
	}

	return TRUE;
}



CString GUtil::s_INIString;
ANE_PTR GUtil::s_ANEHandle = 0;

void
GUtil::ReplaceIni(CString& str)
{
	long start = 0;
	long end = 0;
	while (start < s_INIString.GetLength())
	{
		end = s_INIString.Mid(start).Find('\n');
		if (end < 0)
		{
			end = s_INIString.GetLength() - start;
		}
		CString line = s_INIString.Mid(start, end);
		
		long equal = line.Find('=');
		
		if (equal >= 0)
		{
			CString from = line.Left(equal);
			CString to = line.Mid(equal+1);

			// replace '\n'
			do
			{
				long nlPos = to.Find("\\n");
				if (nlPos<0)
					break;
				to = to.Mid(0,nlPos) + '\n' + to.Mid(nlPos+2);
			} while(1);

			from.TrimLeft(); from.TrimRight();
			to.TrimLeft(); to.TrimRight();

			SReplaceString(str, from, to);
		}

		start = start + end+1;
	}

} // GUtil::ReplaceIni

void 
GUtil::ProcessEvents()
{
	if (s_ANEHandle)
		ANE_ProcessEvents(s_ANEHandle);

} // GUtil::ProcessEvents

//==============================================================================
// SUTRA Project Information Block definition
//==============================================================================

//------------------------------------------------------------------------------
// SUTRAProjectInfo::SUTRAProjectInfo
//------------------------------------------------------------------------------
// Constructor of project-information block
// Fill information block with default values
//------------------------------------------------------------------------------

SUTRAProjectInfo::SUTRAProjectInfo()
{

	valString = "";

	// Pane1:

	/*long*/		NTOBS		= 0;
	/*CString*/		SIMULA		= "";
	/*CString*/		TITLE1		= "";
	/*CString*/		TITLE2		= "";

	// Pane2:

	/*double*/		GNUP		= 0;
	/*double*/		GNUU		= 0;
	/*int*/			IREAD		= 0;
	/*int*/			ISSFLO		= 0;
	/*int*/			ISSTRA		= 0;
	/*long*/		ISTORE		= 0;
	/*int*/			IUNSAT		= 0;
	/*double*/		UP			= 0;

	// Pane3:

	/*double*/		DELT		= 0;
	/*double*/		DTMAX		= 0;
	/*long*/		ITCYC		= 0;
	/*long*/		ITMAX		= 0;
	/*long*/		NPCYC		= 0;
	/*long*/		NUCYC		= 0;
	/*double*/		DTMULT		= 0;
	/*double*/		TMAX		= 0;
	/*double*/		TSTART		= 0;

	// Pane4:

	/*BOOL*/		KBUDG		= 0;
	/*BOOL*/		KELMNT		= 0;
	/*BOOL*/		KINCID		= 0;
	/*BOOL*/		KNODAL		= 0;
	/*BOOL*/		KPLOTP		= 0;
	/*BOOL*/		KPLOTU		= 0;
	/*BOOL*/		KVEL		= 0;
	/*long*/		NPRINT		= 0;
	/*long*/		NOBCYC		= 0;

	// Pane5:

	/*long*/		ITRMAX		= 0;
	/*double*/		RPMAX		= 0;
	/*double*/		RUMAX		= 0;

	// Pane6:

	/*double*/		COMPFL		= 0;
	/*double*/		CW			= 0;
	/*double*/		DRWDU		= 0;
	/*double*/		RHOW0		= 0;
	/*double*/		SIGMAW		= 0;
	/*double*/		URHOW0		= 0;
	/*double*/		VISCO		= 0;

	// Pane7:

	/*CString*/		ADSMOD		= "";
	/*double*/		CHI1		= 0;
	/*double*/		CHI2		= 0;
	/*double*/		COMPMA		= 0;
	/*double*/		CS			= 0;
	/*double*/		RHOS		= 0;
	/*double*/		SIGMAS		= 0;

	// Pane8:

	/*double*/		GRAVX		= 0;
	/*double*/		GRAVY		= 0;
	/*double*/		PRODF0		= 0;
	/*double*/		PRODF1		= 0;
	/*double*/		PRODS0		= 0;
	/*double*/		PRODS1		= 0;

	// Pane9:

	/*int*/			IDIREC		= 0;
	/*long*/		NCHAPI		= 0;
	/*long*/		NCHAPL		= 0;
	/*long*/		NLINPI		= 0;
	/*double*/		PBASE		= 0;
	/*double*/		UBASE		= 0;

	//



}


//------------------------------------------------------------------------------
// SUTRAProjectInfo::~SUTRAProjectInfo
//------------------------------------------------------------------------------
// Distructor
// Clean up before killing project-information block
//------------------------------------------------------------------------------

SUTRAProjectInfo::~SUTRAProjectInfo()
{
	// delete all pointers
}

static CString sDefaultVAL = 
"@NTOBS@\n\
@SIMULA@ SUTRA SOLUTE TRANSPORT\n\
@TITLE1@ Argus Numerical Environments Sample SUTRA Defaults\n\
@TITLE2@ USGS stuff\n\
@GNUP@ 0.1\n\
@GNUU@ 1.0\n\
@IREAD@ 1\n\
@ISSFLO@ 1\n\
@ISSTRA@ 1\n\
@ISTORE@ 9999\n\
@IUNSAT@ \n\
@UP@\n"

"		// Pane3:\n\
@DELT@ 1\n\
@DTMAX@ 1\n\
@ITCYC@ 9999\n\
@ITMAX@ 1\n\
@NPCYC@ 1\n\
@NUCYC@ 1\n\
@DTMULT@ 1.0\n\
@TMAX@ 1.0\n"

"		// Pane4:\n\
@KBUDG@ 1\n\
@KELMNT@\n\
@KINCID@\n\
@KNODAL@\n\
@KPLOTP@\n\
@KPLOTU@\n\
@KVEL@ 1\n\
@NPRINT@ 9999\n"

"		// Pane5:\n\
@ITRMAX@ 1\n\
@RPMAX@		\n\
@RUMAX@		\n"

"		// Pane6:\n\
@COMPFL@\n\
@CW@		\n\
@DRWDU@	\n\
@RHOW0@ 1\n\
@SIGMAW@\n\
@URHOW0@\n\
@VISCO@ 1\n"

"		// Pane7:\n\
@ADSMOD@ NONE      \n\
@CHI1@	\n\
@CHI2@\n\
@COMPMA@	\n\
@CS@\n\
@RHOS@	\n\
@SIGMAS@\n"

"		// Pane8:\n\
@GRAVX@\n\
@GRAVY@	\n\
@PRODF0@	\n\
@PRODF1@	\n\
@PRODS0@	\n\
@PRODS1@	\n"

"		// Pane9:\n\
@IDIRECT@;	\n\
@NCHAPI@		\n\
@NCHAPL@		\n\
@NLINPI@		\n\
@PBASE@		\n\
@UBASE@		\n"

"		// Pane10\n\
@NOBCYC@\n\
@TSTART@	\n";


//------------------------------------------------------------------------------
// SUTRAProjectInfo::PrepareValString
//------------------------------------------------------------------------------
// Initialize the val string (which is used in InitFromValFile) from the .INI file,
// based on the object state.
// Call this member only after you have set the state of the info object.
// 
//------------------------------------------------------------------------------
void 
SUTRAProjectInfo::PrepareValString()
{

	if (valString.IsEmpty())
	{

		CString iniInfo = "";

		CString temp;

		if (this->defaultSetting)
		{
				SReadIniSection("SUTRA.INI", "GENERAL", temp);
				iniInfo += temp;
		}
		else
		{
			
			switch (this->orientation)
			{
				case SUTRAProjectInfo::kAREAL:
						SReadIniSection("SUTRA.INI", "AREAL", temp);
						iniInfo += temp;
					break;

				case SUTRAProjectInfo::kCROSS_SECTIONAL:
						SReadIniSection("SUTRA.INI", "CROSS_SECTIONAL", temp);
						iniInfo += temp;
					break;

				default:
					ASSERT(0);
			}
			
			switch (this->flowConditions)
			{
				case SUTRAProjectInfo::kSATURATED:
						SReadIniSection("SUTRA.INI", "SATURATED", temp);
						iniInfo += temp;
					break;

				case SUTRAProjectInfo::kSAT_UNSAT:
						SReadIniSection("SUTRA.INI", "SAT_UNSAT", temp);
						iniInfo += temp;
					break;

				default:
					ASSERT(0);
			}
			
			switch (this->transportConditions)
			{
				case SUTRAProjectInfo::kSOLUTE_VAR:
						SReadIniSection("SUTRA.INI", "SOLUTE-VAR", temp);
						iniInfo += temp;
					break;

				case SUTRAProjectInfo::kSOLUTE_CONST:
						SReadIniSection("SUTRA.INI", "SOLUTE-CONST", temp);
						iniInfo += temp;
					break;

				case SUTRAProjectInfo::kENERGY:
						SReadIniSection("SUTRA.INI", "ENERGY", temp);
						iniInfo += temp;
					break;

				default:
					ASSERT(0);

			}

			switch (this->modelThickness)
			{
				case SUTRAProjectInfo::kUSER_SPEC:
						SReadIniSection("SUTRA.INI", "USER_SPEC", temp);
						iniInfo += temp;
					break;

				case SUTRAProjectInfo::kCYLINDRICAL:
						SReadIniSection("SUTRA.INI", "CYLINDRICAL", temp);
						iniInfo += temp;
					break;

				default:
					ASSERT(0);
			}
			
			switch (this->meshDesign)
			{
				case SUTRAProjectInfo::kFISHNET_MESH:
						SReadIniSection("SUTRA.INI", "FISHNET_MESH", temp);
						iniInfo += temp;
					break;

				case SUTRAProjectInfo::kIRREGULAR_MESH:
						SReadIniSection("SUTRA.INI", "IRREGULAR_MESH", temp);
						iniInfo += temp;
					break;

				default:
					ASSERT(0);
			}
			


		}
		//

		valString = iniInfo;
	}

	GUtil::SetIniString(valString);
}
//------------------------------------------------------------------------------
// SUTRAProjectInfo::InitFromValFile
//------------------------------------------------------------------------------
// Initialize the object from .VAL file (if available) and using iniInfo
// 
//------------------------------------------------------------------------------
void 
SUTRAProjectInfo::InitFromValFile()
{

	this->PrepareValString();

	CString defaultValues; 
	if (SReadFromFile("SUTRA.VAL", defaultValues))
	{
		defaultValues = valString+defaultValues;
	}
	else
	{
		defaultValues = valString + sDefaultVAL;
	}

	// Pane1:

	SReadDefLong(defaultValues,"@NTOBS@", NTOBS);
	SReadDefString(defaultValues,"@SIMULA@", SIMULA);	
	SReadDefString(defaultValues,"@TITLE1@", TITLE1);	
	SReadDefString(defaultValues,"@TITLE2@", TITLE2);	

	// Pane2:

	SReadDefDouble(defaultValues,"@GNUP@", GNUP);	
	SReadDefDouble(defaultValues,"@GNUU@", GNUU);	
	SReadDefInt(defaultValues,"@IREAD@", IREAD);
	SReadDefInt(defaultValues,"@ISSFLO@", ISSFLO);
	SReadDefInt(defaultValues,"@ISSTRA@", ISSTRA);
	SReadDefLong(defaultValues,"@ISTORE@", ISTORE);
	SReadDefInt(defaultValues,"@IUNSAT@", IUNSAT);
	SReadDefDouble(defaultValues,"@UP@", UP);	

	// Pane3:

	SReadDefDouble(defaultValues,"@DELT@", DELT);	
	SReadDefDouble(defaultValues,"@DTMAX@", DTMAX);	
	SReadDefLong(defaultValues,"@ITCYC@", ITCYC);		
	SReadDefLong(defaultValues,"@ITMAX@", ITMAX);		
	SReadDefLong(defaultValues,"@NPCYC@", NPCYC);		
	SReadDefLong(defaultValues,"@NUCYC@", NUCYC);		
	SReadDefDouble(defaultValues,"@DTMULT@",DTMULT);	
	SReadDefDouble(defaultValues,"@TMAX@", TMAX);		

	// Pane4:

	SReadDefBool(defaultValues,"@KBUDG@", KBUDG);	
	SReadDefBool(defaultValues,"@KELMNT@", KELMNT);	
	SReadDefBool(defaultValues,"@KINCID@", KINCID);	
	SReadDefBool(defaultValues,"@KNODAL@", KNODAL);	
	SReadDefBool(defaultValues,"@KPLOTP@", KPLOTP);	
	SReadDefBool(defaultValues,"@KPLOTU@", KPLOTU);	
	SReadDefBool(defaultValues,"@KVEL@", KVEL);		
	SReadDefLong(defaultValues,"@NPRINT@", NPRINT);	

	// Pane5:

	SReadDefLong(defaultValues,"@ITRMAX@", ITRMAX);		
	SReadDefDouble(defaultValues,"@RPMAX@", RPMAX);		
	SReadDefDouble(defaultValues,"@RUMAX@", RUMAX);		

	// Pane6:

	SReadDefDouble(defaultValues,"@COMPFL@", COMPFL);
	SReadDefDouble(defaultValues,"@CW@", CW);		
	SReadDefDouble(defaultValues,"@DRWDU@", DRWDU);	
	SReadDefDouble(defaultValues,"@RHOW0@", RHOW0);	
	SReadDefDouble(defaultValues,"@SIGMAW@", SIGMAW);
	SReadDefDouble(defaultValues,"@URHOW0@", URHOW0);
	SReadDefDouble(defaultValues,"@VISCO@", VISCO);	

	// Pane7:

	SReadDefString(defaultValues,"@ADSMOD@", ADSMOD);	
	SReadDefDouble(defaultValues,"@CHI1@", CHI1);		
	SReadDefDouble(defaultValues,"@CHI2@", CHI2);		
	SReadDefDouble(defaultValues,"@COMPMA@", COMPMA);	
	SReadDefDouble(defaultValues,"@CS@", CS);			
	SReadDefDouble(defaultValues,"@RHOS@", RHOS);		
	SReadDefDouble(defaultValues,"@SIGMAS@", SIGMAS);	

	// Pane8:

	SReadDefDouble(defaultValues,"@GRAVX@", GRAVX);		
	SReadDefDouble(defaultValues,"@GRAVY@", GRAVY);		
	SReadDefDouble(defaultValues,"@PRODF0@", PRODF0);	
	SReadDefDouble(defaultValues,"@PRODF1@", PRODF1);	
	SReadDefDouble(defaultValues,"@PRODS0@", PRODS0);	
	SReadDefDouble(defaultValues,"@PRODS1@", PRODS1);	

	// Pane9:

	SReadDefInt(defaultValues,"@IDIREC@", IDIREC);	
	SReadDefLong(defaultValues,"@NCHAPI@", NCHAPI);		
	SReadDefLong(defaultValues,"@NCHAPL@", NCHAPL);		
	SReadDefLong(defaultValues,"@NLINPI@", NLINPI);		
	SReadDefDouble(defaultValues,"@PBASE@", PBASE);		
	SReadDefDouble(defaultValues,"@UBASE@", UBASE);		

	// Pane10

	SReadDefLong(defaultValues,"@NOBCYC@", NOBCYC);		
	SReadDefDouble(defaultValues,"@TSTART@", TSTART);	

}



//------------------------------------------------------------------------------
// SUTRAProjectInfo::Save
//------------------------------------------------------------------------------
// Called by MeshMaker to save project-information block
//------------------------------------------------------------------------------
void
SUTRAProjectInfo::Save(CString& saveTo)
{
	// add all layer info to the saveTo CString
	char buf[512]; // !!! CAUTION: buffer may overflow

	// version information
	saveTo += "USGS-SUTRA 3 ";	// mark product and release

	// user specified model information
	sprintf(buf, "%ld ", defaultSetting);	saveTo += buf;
	if (!defaultSetting)
	{
		sprintf(buf, "%ld ", flowConditions);	saveTo += buf;
		sprintf(buf, "%ld ", meshDesign);	saveTo += buf;
		sprintf(buf, "%ld ", modelThickness);	saveTo += buf;
		sprintf(buf, "%ld ", orientation);	saveTo += buf;
		sprintf(buf, "%ld ", transportConditions);	saveTo += buf;
	}

	// Pane1:

	sprintf(buf, "%ld ", NTOBS);	saveTo += buf;

	sprintf(buf, "%ld ", strlen(SIMULA)); saveTo += buf;
	sprintf(buf, "%s", SIMULA);	saveTo += buf;
	sprintf(buf, "%ld ", strlen(TITLE1)); saveTo += buf;
	sprintf(buf, "%s", TITLE1);	saveTo += buf;
	
	sprintf(buf, "%ld ", strlen(TITLE2)); saveTo += buf;
	sprintf(buf, "%s", TITLE2);	saveTo += buf;
	
	// Pane2:

	sprintf(buf, "%.15lg ", GNUP);		saveTo += buf;
	sprintf(buf, "%.15lg ", GNUU);		saveTo += buf;
	sprintf(buf, "%d ", IREAD);		saveTo += buf;
	sprintf(buf, "%d ", ISSFLO);	saveTo += buf;
	sprintf(buf, "%d ", ISSTRA);	saveTo += buf;
	sprintf(buf, "%ld ", ISTORE);	saveTo += buf;
	sprintf(buf, "%d ", IUNSAT);	saveTo += buf;
	sprintf(buf, "%.15lg ", UP);		saveTo += buf;

	// Pane3:

	sprintf(buf, "%.15lg ", DELT);		saveTo += buf;
	sprintf(buf, "%.15lg ", DTMAX);	saveTo += buf;
	sprintf(buf, "%ld ", ITCYC);	saveTo += buf;
	sprintf(buf, "%ld ", ITMAX);	saveTo += buf;
	sprintf(buf, "%ld ", NPCYC);	saveTo += buf;
	sprintf(buf, "%ld ", NUCYC);	saveTo += buf;
	sprintf(buf, "%.15lg ", DTMULT);	saveTo += buf;
	sprintf(buf, "%.15lg ", TMAX);		saveTo += buf;

	// Pane4:

	sprintf(buf, "%d ", KBUDG);		saveTo += buf;
	sprintf(buf, "%d ", KELMNT);	saveTo += buf;
	sprintf(buf, "%d ", KINCID);	saveTo += buf;
	sprintf(buf, "%d ", KNODAL);	saveTo += buf;
	sprintf(buf, "%d ", KPLOTP);	saveTo += buf;
	sprintf(buf, "%d ", KPLOTU);	saveTo += buf;
	sprintf(buf, "%d ", KVEL);		saveTo += buf;
	sprintf(buf, "%d ", NPRINT);	saveTo += buf;

	// Pane5:

	sprintf(buf, "%ld ", ITRMAX);	saveTo += buf;
	sprintf(buf, "%.15lg ", RPMAX);	saveTo += buf;
	sprintf(buf, "%.15lg ", RUMAX);	saveTo += buf;

	// Pane6:

	sprintf(buf, "%.15lg ", COMPFL);	saveTo += buf;
	sprintf(buf, "%.15lg ", CW);	saveTo += buf;
	sprintf(buf, "%.15lg ", DRWDU);	saveTo += buf;
	sprintf(buf, "%.15lg ", RHOW0);	saveTo += buf;
	sprintf(buf, "%.15lg ", SIGMAW);	saveTo += buf;
	sprintf(buf, "%.15lg ", URHOW0);	saveTo += buf;
	sprintf(buf, "%.15lg ", VISCO);	saveTo += buf;

	// Pane7:

	sprintf(buf, "%ld ", strlen(ADSMOD)); saveTo += buf;
	sprintf(buf, "%s", ADSMOD);	saveTo += buf;
	sprintf(buf, "%.15lg ", CHI1);		saveTo += buf;
	sprintf(buf, "%.15lg ", CHI2);		saveTo += buf;
	sprintf(buf, "%.15lg ", COMPMA);	saveTo += buf;
	sprintf(buf, "%.15lg ", CS);		saveTo += buf;
	sprintf(buf, "%.15lg ", RHOS);		saveTo += buf;
	sprintf(buf, "%.15lg ", SIGMAS);	saveTo += buf;

	// Pane8:

	sprintf(buf, "%.15lg ", GRAVX);	saveTo += buf;
	sprintf(buf, "%.15lg ", GRAVY);	saveTo += buf;
	sprintf(buf, "%.15lg ", PRODF0);	saveTo += buf;
	sprintf(buf, "%.15lg ", PRODF1);	saveTo += buf;
	sprintf(buf, "%.15lg ", PRODS0);	saveTo += buf;
	sprintf(buf, "%.15lg ", PRODS1);	saveTo += buf;

	// Pane9:

	sprintf(buf, "%d ", IDIREC);	saveTo += buf;
	sprintf(buf, "%ld ", NCHAPI);	saveTo += buf;
	sprintf(buf, "%ld ", NCHAPL);	saveTo += buf;
	sprintf(buf, "%ld ", NLINPI);	saveTo += buf;
	sprintf(buf, "%.15lg ", PBASE);	saveTo += buf;
	sprintf(buf, "%.15lg ", UBASE);	saveTo += buf;

	// Pane10

	sprintf(buf, "%ld ", NOBCYC);	saveTo += buf;
	sprintf(buf, "%.15lg ", TSTART);	saveTo += buf;
}


//------------------------------------------------------------------------------
// SUTRAProjectInfo::Load
//------------------------------------------------------------------------------
// Called by MeshMaker to load project-information block
//------------------------------------------------------------------------------

void
SUTRAProjectInfo::Load(ANE_PTR aneHandle, const char* loadFrom, int* position)
{

	int n;
	int p = 0;
	int tmpp = 0;
	long l;

	// Version Information
	int version;
	
	if (!strncmp("USGS-SUTRA", loadFrom, 9))
	{
		p = 10;
		n = sscanf(loadFrom + p, "%ld%n", &version,	&tmpp);	p += tmpp; ASSERT(n==1);
	}
	else
	{
		// unknown file version
		version = 0;
	} 

	if (version > 1)
	{
		// user specified model information
		n = sscanf(loadFrom + p, "%ld%n", &defaultSetting,	&tmpp);	p += tmpp; ASSERT(n==1);
		if (!defaultSetting)
		{
			n = sscanf(loadFrom + p, "%ld%n", &flowConditions,	&tmpp);	p += tmpp; ASSERT(n==1);
			n = sscanf(loadFrom + p, "%ld%n", &meshDesign,	&tmpp);	p += tmpp; ASSERT(n==1);
			n = sscanf(loadFrom + p, "%ld%n", &modelThickness,	&tmpp);	p += tmpp; ASSERT(n==1);
			n = sscanf(loadFrom + p, "%ld%n", &orientation,	&tmpp);	p += tmpp; ASSERT(n==1);
			n = sscanf(loadFrom + p, "%ld%n", &transportConditions,	&tmpp);	p += tmpp; ASSERT(n==1);
		}
	}
	else
	{
		ANE_UserMessageOk(aneHandle, "This USGS SUTRA PIE Project was created with an outdated version. It's will probably won't work properly with the current version (version 2).");
	}
	// Pane1:

	n = sscanf(loadFrom + p, "%ld%n", &NTOBS,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%ld%n", &l,	&tmpp);	p += tmpp; ASSERT(n==1);
	SIMULA = CString(loadFrom+p+1, l); p+=l+1;
	n = sscanf(loadFrom + p, "%ld%n", &l,	&tmpp);	p += tmpp; ASSERT(n==1);
	TITLE1 = CString(loadFrom+p+1, l); p+=l+1;

	if (version > 2)
	{
		n = sscanf(loadFrom + p, "%ld%n", &l,	&tmpp);	p += tmpp; ASSERT(n==1);
		TITLE2 = CString(loadFrom+p+1, l); p+=l+1;
	}
	else
	{
		TITLE1 = "";
	}
	// Pane2:

	n = sscanf(loadFrom + p, "%lf%n", &GNUP,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &GNUU,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%d%n", &IREAD,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%d%n", &ISSFLO,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%d%n", &ISSTRA,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%ld%n", &ISTORE,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%d%n", &IUNSAT,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &UP,		&tmpp);	p += tmpp; ASSERT(n==1);

	// Pane3
	n = sscanf(loadFrom + p, "%lf%n", &DELT,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &DTMAX,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%ld%n", &ITCYC,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%ld%n", &ITMAX,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%ld%n", &NPCYC,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%ld%n", &NUCYC,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &DTMULT,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &TMAX,	&tmpp);	p += tmpp; ASSERT(n==1);

	// Pane4:

	n = sscanf(loadFrom + p, "%d%n", &KBUDG,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%d%n", &KELMNT,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%d%n", &KINCID,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%d%n", &KNODAL,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%d%n", &KPLOTP,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%d%n", &KPLOTU,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%d%n", &KVEL,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%d%n", &NPRINT,	&tmpp);	p += tmpp; ASSERT(n==1);

	// Pane5:

	n = sscanf(loadFrom + p, "%ld%n", &ITRMAX,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &RPMAX,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &RUMAX,	&tmpp);	p += tmpp; ASSERT(n==1);

	// Pane6:
	n = sscanf(loadFrom + p, "%lf%n", &COMPFL,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &CW,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &DRWDU,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &RHOW0,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &SIGMAW,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &URHOW0,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &VISCO,	&tmpp);	p += tmpp; ASSERT(n==1);
	
	// Pane7:

	n = sscanf(loadFrom + p, "%ld%n", &l,	&tmpp);	p += tmpp; ASSERT(n==1);
	ADSMOD = CString(loadFrom+p+1, l); p+=l+1;
	n = sscanf(loadFrom + p, "%lf%n", &CHI1,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &CHI2,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &COMPMA,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &CS,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &RHOS,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &SIGMAS,	&tmpp);	p += tmpp; ASSERT(n==1);

	// Pane8:

	n = sscanf(loadFrom + p, "%lf%n", &GRAVX,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &GRAVY,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &PRODF0,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &PRODF1,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &PRODS0,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &PRODS1,	&tmpp);	p += tmpp; ASSERT(n==1);
	
	// Pane9:
	
	n = sscanf(loadFrom + p, "%d%n", &IDIREC,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%ld%n", &NCHAPI,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%ld%n", &NCHAPL,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%ld%n", &NLINPI,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &PBASE,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &UBASE,	&tmpp);	p += tmpp; ASSERT(n==1);

	// Pane10

	n = sscanf(loadFrom + p, "%ld%n", &NOBCYC,	&tmpp);	p += tmpp; ASSERT(n==1);
	n = sscanf(loadFrom + p, "%lf%n", &TSTART,	&tmpp);	p += tmpp; ASSERT(n==1);
	
	*position = p;

}


//==============================================================================
// PIE Project definition
//==============================================================================

//------------------------------------------------------------------------------
// GSUTRAProjectNew:
//------------------------------------------------------------------------------
// Called by MeshMaker when the user selects New SUTRA Project
// Show dialog and fill project-information block
//------------------------------------------------------------------------------
ANE_BOOL GSUTRAProjectNew(ANE_PTR aneHandle, ANE_PTR* rPIEHandle,  ANE_STR*  returnLayerTemplate)
{
	TRY
	{
		GUtil::SetANEHandle(aneHandle);

		// display copyright message
		CBaseDialog copyright(IDD_COPYRIGHT);
		copyright.DoModal();

		// load SUTRA MODEL SETUP dialog
		CModelSetupDialog modelSetup;
		if (modelSetup.DoModal() != IDOK)
		{
			GUtil::SetANEHandle(0);
			return FALSE;
		}

		SUTRAProjectInfo* info = new SUTRAProjectInfo();

		modelSetup.FillInfo(info);

		info->InitFromValFile();

		// prepare dialog box

		if (info->ShowDialog())
		{

//========================================================
// Build layer structure string
//========================================================

		CString s;

		CString layersTemplate; 
		if (!SReadFromFile("SUTRA.LAY", layersTemplate))
			layersTemplate =
"Layer:\n\
{\n\
	Name: \"SUTRA Mesh\"\n\
	Units: \"\"\n\
	Type: \"QuadMesh\"\n\
	Visible: Yes\n\
	Domain Layer: \"Domain Outline\"\n\
	Density Layer: \"@Mesh Density@\"\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"THICK\"\n\
		Units: \"L\"\n\
		Value Type: Real\n\
		Value: \"@THICK_VALUE@\"\n\
		Parameter Type: Node\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"POR\"\n\
		Units: \"1\"\n\
		Value Type: Real\n\
		Value: \"Porosity\"\n\
		Parameter Type: Node\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"PMAX\"\n\
		Units: \"@max_min_units@\"\n\
		Value Type: Real\n\
		Value: \"@Permeability@.maximum\"\n\
		Parameter Type: Element\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"PMIN\"\n\
		Units: \"@max_min_units@\"\n\
		Value Type: Real\n\
		Value: \"@Permeability@.minimum\"\n\
		Parameter Type: Element\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"ANGLEX\"\n\
		Units: \"degrees\"\n\
		Value Type: Real\n\
		Value: \"@Permeability@.angle_of_max_to_x_axis\"\n\
		Parameter Type: Element\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"ALMAX\"\n\
		Units: \"L\"\n\
		Value Type: Real\n\
		Value: \"Dispersivity.longdisp_in_max_permdir\"\n\
		Parameter Type: Element\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"ALMIN\"\n\
		Units: \"L\"\n\
		Value Type: Real\n\
		Value: \"Dispersivity.longdisp_in_min_permdir\"\n\
		Parameter Type: Element\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"ATMAX\"\n\
		Units: \"L\"\n\
		Value Type: Real\n\
		Value: \"Dispersivity.trandisp_in_max_permdir\"\n\
		Parameter Type: Element\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"ATMIN\"\n\
		Units: \"L\"\n\
		Value Type: Real\n\
		Value: \"Dispersivity.trandisp_in_min_permdir\"\n\
		Parameter Type: Element\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"QIN\"\n\
		Units: \"@fluid_source_units@\"\n\
		Value Type: Real\n\
		Value: \"If(NodeAboveCntr(Sources of Fluid)=0,$n/a, If(NodeAboveCntr(Sources of Fluid)=1, Sources of Fluid.RESULTANT_FLUID_SOURCE, If(NodeAboveCntr(Sources of Fluid)=2, Sources of Fluid.RESULTANT_FLUID_SOURCE*NodeEffectiveLength(Sources of Fluid), NodeEffectiveArea(Sources of Fluid.RESULTANT_FLUID_SOURCE))))\"\n\
		Parameter Type: Node\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"UIN\"\n\
		Units: \"@C or T@\"\n\
		Value Type: Real\n\
		Value: \"Sources of Fluid.@conc_or_temp@_of_source\"\n\
		Parameter Type: Node\n\
	}\n\
	\n"

"	Parameter: \n\
	{\n\
		Name: \"QUIN\"\n\
		Units: \"@sources_of_e_or_s_units@/s\"\n\
		Value Type: Real\n\
		Value: \"If(NodeAboveCntr(Sources of @Energy or Solute@)=0, $n/a, If(NodeAboveCntr(Sources of @Energy or Solute@)=1, Sources of @Energy or Solute@.RESULTANT_@ENERGY_OR_SOLUTE@_SOURCES, If(NodeAboveCntr(Sources of @Energy or Solute@)=2, Sources of @Energy or Solute@.RESULTANT_@ENERGY_OR_SOLUTE@_SOURCES*NodeEffectiveLength(Sources of @Energy or Solute@), NodeEffectiveArea(Sources of @Energy or Solute@.RESULTANT_@ENERGY_OR_SOLUTE@_SOURCES))))\"\n\
		Parameter Type: Node\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"PBC\"\n\
		Units: \"@spec_pressure_units@\"\n\
		Value Type: Real\n\
		Value: \"Specified @Pressure@.specified_@pressure@\"\n\
		Parameter Type: Node\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"pUBC\"\n\
		Units: \"@C or T@\"\n\
		Value Type: Real\n\
		Value: \"Specified @Pressure@.@conc_or_temp@\"\n\
		Parameter Type: Node\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"UBC\"\n\
		Units: \"@C or T@\"\n\
		Value Type: Real\n\
		Value: \"Specified @Conc or Temp@.specified_@conc_or_temp@\"\n\
		Parameter Type: Node\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"PVEC\"\n\
		Units: \"@spec_pressure_units@\"\n\
		Value Type: Real\n\
		Value: \"Initial @Pressure@.initial_@pressure@\"\n\
		Parameter Type: Node\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"UVEC\"\n\
		Units: \"@C or T@\"\n\
		Value Type: Real\n\
		Value: \"Initial @Conc or Temp@.initial_@conc_or_temp@\"\n\
		Parameter Type: Node\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"INOB\"\n\
		Units: \"\"\n\
		Value Type: Integer\n\
		#Value: \"If(Observation.is_observed,NodeNumber(),$n/a)\"\n\
		Value: \"If(Observation.is_observed,NodeNumber(),0)\"\n\
		Parameter Type: Node\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"BC_VALIDITY\"\n\
		Units: \"\"\n\
		Value Type: String\n\
		Value: \"If((If(isna(Specified @Pressure@.specified_@pressure@),0,1)+If(isna(Specified @Conc or Temp@.specified_@conc_or_temp@),0,1)+If(isna(Sources of @Energy or Solute@.RESULTANT_@ENERGY_OR_SOLUTE@_SOURCES),0,1)+If(isna(Sources of Fluid.RESULTANT_FLUID_SOURCE),0,1))>1,\\\"more than 1 BC\\\",\\\"OK\\\")\"\n\
		Parameter Type: Node\n\
	}\n\
	\n\
	Template: \n\
	{\n\
		Line: \"# editable SUTRA template goes here\"\n\
	}\n\
}\n"


"@FISHNET MESH LAYOUT LAYER@:\n\
{\n\
	Name: \"FishNet Mesh Layout\"\n\
	Units: \"\"\n\
	Type: \"Information\"\n\
	Visible: Yes\n\
	Interpretation Method: Nearest\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"elements_in_x\"\n\
		Units: \"\"\n\
		Value Type: Integer\n\
		Value: \"$n/a\"\n\
		Parameter Type: Layer\n\
	}\n"

"	Parameter: \n\
	{\n\
		Name: \"elements_in_y\"\n\
		Units: \"\"\n\
		Value Type: Integer\n\
		Value: \"$n/a\"\n\
		Parameter Type: Layer\n\
	}\n\
}\n"


"Layer:\n\
{\n\
	Name: \"Domain Outline\"\n\
	Units: \"\"\n\
	Type: \"Domain\"\n\
	Visible: Yes\n\
	Interpretation Method: Nearest\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"Domain Outline\"\n\
		Units: \"\"\n\
		Value Type: Real\n\
		Value: \"0\"\n\
		Parameter Type: Layer\n\
	}\n\
}\n"


"@MESH DENSITY LAYER@:\n\
{\n\
	Name: \"Mesh Density\"\n\
	Units: \"\"\n\
	Type: \"Information\"\n\
	Visible: No\n\
	Interpretation Method: Nearest\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"Mesh Density\"\n\
		Units: \"\"\n\
		Value Type: Real\n\
		Value: \"0\"\n\
		Parameter Type: Layer\n\
	}\n\
}\n"


"@THICKNESS LAYER@:\n\
{\n\
	Name: \"Thickness\"\n\
	Units: \"\"\n\
	Type: \"Information\"\n\
	Visible: No\n\
	Interpretation Method: Interpolate\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"thickness\"\n\
		Units: \"L\"\n\
		Value Type: Real\n\
		Value: \"@thickness value@\"\n\
		Parameter Type: Layer\n\
	}\n\
}\n"


"Layer:\n\
{\n\
	Name: \"Porosity\"\n\
	Units: \"\"\n\
	Type: \"Information\"\n\
	Visible: No\n\
	Interpretation Method: Nearest\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"porosity\"\n\
		Units: \"1\"\n\
		Value Type: Real\n\
		Value: \"0.1\"\n\
		Parameter Type: Layer\n\
	}\n\
}\n"


"Layer:\n\
{\n\
	Name: \"@Permeability@\"\n\
	Units: \"\"\n\
	Type: \"Information\"\n\
	Visible: No\n\
	Interpretation Method: Nearest\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"maximum\"\n\
		Units: \"@max_min_units@\"\n\
		Value Type: Real\n\
		Value: \"1.0E-3\"\n\
		Parameter Type: Layer\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"minimum\"\n\
		Units: \"@max_min_units@\"\n\
		Value Type: Real\n\
		Value: \"1.0E-3\"\n\
		Parameter Type: Layer\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"angle_of_max_to_x_axis\"\n\
		Units: \"degrees\"\n\
		Value Type: Real\n\
		Value: \"0\"\n\
		Parameter Type: Layer\n\
	}\n\
}\n"


"Layer:\n\
{\n\
	Name: \"Dispersivity\"\n\
	Units: \"\"\n\
	Type: \"Information\"\n\
	Visible: No\n\
	Interpretation Method: Nearest\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"longdisp_in_max_permdir\"\n\
		Units: \"L\"\n\
		Value Type: Real\n\
		Value: \"0.5\"\n\
		Parameter Type: Layer\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"longdisp_in_min_permdir\"\n\
		Units: \"L\"\n\
		Value Type: Real\n\
		Value: \"0.5\"\n\
		Parameter Type: Layer\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"trandisp_in_max_permdir\"\n\
		Units: \"L\"\n\
		Value Type: Real\n\
		Value: \"0.1\"\n\
		Parameter Type: Layer\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"trandisp_in_min_permdir\"\n\
		Units: \"L\"\n\
		Value Type: Real\n\
		Value: \"0.1\"\n\
		Parameter Type: Layer\n\
	}\n\
}\n"


"@UNSATURATED PROPERTIES LAYER@:\n\
{\n\
	Name: \"Unsaturated Properties\"\n\
	Units: \"\"\n\
	Type: \"Information\"\n\
	Visible: No\n\
	Interpretation Method: Exact\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"region\"\n\
		Units: \"\"\n\
		Value Type: Integer\n\
		Value: \"0\"\n\
		Parameter Type: Layer\n\
	}\n\
}\n"


"Layer:\n\
{\n\
	Name: \"Sources of Fluid\"\n\
	Units: \"\"\n\
	Type: \"Information\"\n\
	Visible: No\n\
	Interpretation Method: Exact\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"total_source\"\n\
		Units: \"@fluid_source_units@\"\n\
		Value Type: Real\n\
		Value: \"$n/a\"\n\
		Parameter Type: Layer\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"specific_source\"\n\
		Units: \"(@fluid_source_units@)/(L or L^2)\"\n\
		Value Type: Real\n\
		Value: \"$n/a\"\n\
		Parameter Type: Layer\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"@conc_or_temp@_of_source\"\n\
		Units: \"@C or degC@\"\n\
		Value Type: Real\n\
		Value: \"$n/a\"\n\
		Parameter Type: Layer\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"RESULTANT_FLUID_SOURCE\"\n\
		Units: \"calculated\"\n\
		Value Type: Real\n\
		Value: \"If(!isna(Sources of Fluid.specific_source)|!isna(Sources of Fluid.total_source), If(!isna(Sources of Fluid.specific_source), Sources of Fluid.specific_source, If(ContourType()=3, Sources of Fluid.total_source/ContourArea(), If(ContourType()=2, Sources of Fluid.total_source/ContourLength(), Sources of Fluid.total_source))), $n/a)\"\n\
		Parameter Type: Layer\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"QA\"\n\
		Units: \"\"\n\
		Value Type: String\n\
		Value: \"If(!isna(Sources of Fluid.specific_source)&!isna(Sources of Fluid),\\\"Badly Defined\\\",\\\"OK\\\")\"\n\
		Parameter Type: Layer\n\
	}\n\
}\n"


"Layer:\n\
{\n\
	Name: \"Sources of @Energy or Solute@\"\n\
	Units: \"\"\n\
	Type: \"Information\"\n\
	Visible: No\n\
	Interpretation Method: Exact\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"total_source\"\n\
		Units: \"@sources_of_e_or_s_units@/s\"\n\
		Value Type: Real\n\
		Value: \"$n/a\"\n\
		Parameter Type: Layer\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"specific_source\"\n\
		Units: \"(@sources_of_e_or_s_units@/s)/(L or L^2)\"\n\
		Value Type: Real\n\
		Value: \"$n/a\"\n\
		Parameter Type: Layer\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"RESULTANT_@ENERGY_OR_SOLUTE@_SOURCES\"\n\
		Units: \"calculated\"\n\
		Value Type: Real\n\
		Value: \"If(!isna(Sources of @Energy or Solute@.specific_source)|!isna(Sources of @Energy or Solute@.total_source), If(!isna(Sources of @Energy or Solute@.specific_source), Sources of @Energy or Solute@.specific_source, If(ContourType()=3, Sources of @Energy or Solute@.total_source/ContourArea(), If(ContourType()=2, Sources of @Energy or Solute@.total_source/ContourLength(), Sources of @Energy or Solute@.total_source))), $n/a)\"\n\
		Parameter Type: Layer\n\
	}\n\
}\n"


"Layer:\n\
{\n\
	Name: \"Specified @Pressure@\"\n\
	Units: \"\"\n\
	Type: \"Information\"\n\
	Visible: No\n\
	Interpretation Method: Exact\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"specified_@pressure@\"\n\
		Units: \"@spec_pressure_units@\"\n\
		Value Type: Real\n\
		Value: \"$n/a\"\n\
		Parameter Type: Layer\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"@conc_or_temp@\"\n\
		Units: \"@C or degC@\"\n\
		Value Type: Real\n\
		Value: \"$n/a\"\n\
		Parameter Type: Layer\n\
	}\n\
}\n"


"Layer:\n\
{\n\
	Name: \"Specified @Conc or Temp@\"\n\
	Units: \"\"\n\
	Type: \"Information\"\n\
	Visible: No\n\
	Interpretation Method: Exact\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"specified_@conc_or_temp@\"\n\
		Units: \"@C or degC@\"\n\
		Value Type: Real\n\
		Value: \"$n/a\"\n\
		Parameter Type: Layer\n\
	}\n\
}\n"


"Layer:\n\
{\n\
	Name: \"Observation\"\n\
	Units: \"\"\n\
	Type: \"Information\"\n\
	Visible: No\n\
	Interpretation Method: Exact\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"is_observed\"\n\
		Units: \"True or False\"\n\
		Value Type: Boolean\n\
		Value: \"0\"\n\
		Parameter Type: Layer\n\
	}\n\
}\n"


"Layer:\n\
{\n\
	Name: \"Initial @Pressure@\"\n\
	Units: \"\"\n\
	Type: \"Information\"\n\
	Visible: No\n\
	Interpretation Method: Nearest\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"initial_@pressure@\"\n\
		Units: \"@spec_pressure_units@\"\n\
		Value Type: Real\n\
		Value: \"0\"\n\
		Parameter Type: Layer\n\
	}\n\
}\n"


"Layer:\n\
{\n\
	Name: \"Initial @Conc or Temp@\"\n\
	Units: \"\"\n\
	Type: \"Information\"\n\
	Visible: No\n\
	Interpretation Method: Nearest\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"initial_@conc_or_temp@\"\n\
		Units: \"@C or degC@\"\n\
		Value Type: Real\n\
		Value: \"0\"\n\
		Parameter Type: Layer\n\
	}\n\
}\n"


"Layer:\n\
{\n\
	Name: \"Map\"\n\
	Units: \"\"\n\
	Type: \"Maps\"\n\
	Visible: Yes\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"Map\"\n\
		Units: \"\"\n\
		Value Type: Real\n\
		Value: \"0\"\n\
		Parameter Type: Layer\n\
	}\n\
}\n"


"Layer:\n\
{\n\
	Name: \"Contour Data\"\n\
	Units: \"Units\"\n\
	Type: \"Data\"\n\
	Visible: No\n\
	Interpretation Method: Nearest\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"@PRESSURE@\"\n\
		Units: \"@spec_pressure_units@\"\n\
		Value Type: Real\n\
		Value: \"0\"\n\
		Parameter Type: Layer\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"@CONC_OR_TEMP@\"\n\
		Units: \"@C or degC@\"\n\
		Value Type: Real\n\
		Value: \"0\"\n\
		Parameter Type: Layer\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"SATURATION\"\n\
		Units: \"1\"\n\
		Value Type: Real\n\
		Value: \"0\"\n\
		Parameter Type: Layer\n\
	}\n\
}\n"


"Layer:\n\
{\n\
	Name: \"Velocity Data\"\n\
	Units: \"Units\"\n\
	Type: \"Data\"\n\
	Visible: No\n\
	Interpretation Method: Nearest\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"velocity_x\"\n\
		Units: \"L/s\"\n\
		Value Type: Real\n\
		Value: \"0\"\n\
		Parameter Type: Layer\n\
	}\n\
	\n\
	Parameter: \n\
	{\n\
		Name: \"velocity_y\"\n\
		Units: \"L/s\"\n\
		Value Type: Real\n\
		Value: \"0\"\n\
		Parameter Type: Layer\n\
	}\n\
}\n"
;



	
//========================================================
// End of layer structure definition
//========================================================


			//
			// prepare a characters buffer to return
			// the template to MeshMaker
			GUtil::ReplaceIni(layersTemplate);


			char* buf = new char[layersTemplate.GetLength()+1];
			strcpy(buf, layersTemplate);

			*rPIEHandle = info;
			*returnLayerTemplate = buf;
	
			GUtil::SetANEHandle(0);
			return TRUE;
		}
		else
		{
			GUtil::SetANEHandle(0);
			return FALSE;
		}
	}
	CATCH_ALL(e)
	{
		GUtil::SetANEHandle(0);
		return FALSE;
	}
	END_CATCH_ALL
}


//------------------------------------------------------------------------------
// GSUTRAProjectEdit:
//------------------------------------------------------------------------------
// Called by MeshMaker when the user selects Edit Project Information
// Show dialog and update project-information block
//------------------------------------------------------------------------------
ANE_BOOL  GSUTRAProjectEdit(ANE_PTR aneHandle, ANE_PTR pieHandle)
{
	TRY
	{
		SUTRAProjectInfo* info = (SUTRAProjectInfo*)pieHandle;

		//
		// prepare dialog box

		return info->ShowDialog();
	}
	CATCH_ALL(e)
	{
		 return FALSE;
	}
	END_CATCH_ALL
}


//------------------------------------------------------------------------------
// GSUTRAProjectClean:
//------------------------------------------------------------------------------
// Called by MeshMaker before quit
// Delete project-information block
//------------------------------------------------------------------------------
void GSUTRAProjectClean(ANE_PTR Handle, ANE_PTR pieHandle)
{
	TRY
	{
		SUTRAProjectInfo* projectInfo= (SUTRAProjectInfo*)pieHandle;
		delete projectInfo;
	}
	CATCH_ALL(e)
	{
		// return FALSE;
	}
	END_CATCH_ALL
}

//------------------------------------------------------------------------------
// GSUTRAProjectSave:
//------------------------------------------------------------------------------
// Called by MeshMaker at save
// Ask project-information block to save it's data
//------------------------------------------------------------------------------
void GSUTRAProjectSave(ANE_PTR Handle, ANE_PTR pieHandle, ANE_STR* rSaveInfo)
{
	TRY
	{
		SUTRAProjectInfo* projectInfo= (SUTRAProjectInfo*)pieHandle;
		CString saveTo;

		projectInfo->Save(saveTo);
		char* returnBuf = new char[saveTo.GetLength()+1];
		strcpy(returnBuf, saveTo);
		*rSaveInfo = returnBuf;
	}
	CATCH_ALL(e)
	{
		rSaveInfo = 0;
		// return FALSE;
	}
	END_CATCH_ALL
}


//------------------------------------------------------------------------------
// GSUTRAProjectLoad:
//------------------------------------------------------------------------------
// Called by MeshMaker at load
// Ask project-information block to read it's data
//------------------------------------------------------------------------------
void GSUTRAProjectLoad(ANE_PTR aneHandle, ANE_PTR* rPIEHandle, const ANE_STR LoadInfo)
{
	TRY
	{
		SUTRAProjectInfo* projectInfo= new SUTRAProjectInfo();

		int position = 0;
		projectInfo->Load(aneHandle, LoadInfo, &position);
		
		*rPIEHandle = projectInfo;
	}
	CATCH_ALL(e)
	{
		rPIEHandle = 0;
		// return FALSE;
	}
	END_CATCH_ALL
}

//==============================================================================
// PIE Function definition
//==============================================================================

//------------------------------------------------------------------------------
// GRunSUTRAMMFun:
//------------------------------------------------------------------------------
// Function definition:
// 	Define a MeshMaker function "SUTRA_RunSUTRA" that calls Windows
//	to run the batch file "RUN_SUTRA.BAT"
//
// This function is used by the SUTRA export template
//
//------------------------------------------------------------------------------

static void
GRunSUTRAMMFun(const ANE_DOUBLE&				refPtX,
				const ANE_DOUBLE&			refPtY,
				ANE_INT16					numParams,
				const ANE_PTR*		parameters,
				ANE_PTR				myHandle,
				ANE_PTR		reply )
{
	char buf[100];


	sprintf(buf, "command.com /c runsutra.bat");

#ifndef _MAC		

	WinExec(buf, SW_SHOWNORMAL);
#endif

	*(ANE_DOUBLE*)reply = 0;

} // GRunSUTRAMMFun

static double sTemp;
static double* sArray = 0;
static long sArraySize = 0;

//------------------------------------------------------------------------------
// GSetTempMMFun:
//------------------------------------------------------------------------------
// Function definition:
// 	Define a MeshMaker function "SUTRA_SetTemp" 
//
// This function is used by the SUTRA export template
//
//------------------------------------------------------------------------------

static void
GSetTempMMFun(const ANE_DOUBLE&				refPtX,
				const ANE_DOUBLE&			refPtY,
				ANE_INT16				numParams,
				const ANE_PTR*		parameters,
				ANE_PTR				myHandle,
				ANE_PTR		reply )
{
	ANE_DOUBLE val = *(ANE_DOUBLE*)(parameters[0]);
	sTemp = val;

	*(ANE_DOUBLE*)reply = 0;

} //


//------------------------------------------------------------------------------
// GGetTempMMFun:
//------------------------------------------------------------------------------
// Function definition:
// 	Define a MeshMaker function "SUTRA_GetTemp" 
//
// This function is used by the SUTRA export template
//
//------------------------------------------------------------------------------

static void
GGetTempMMFun(const ANE_DOUBLE&				refPtX,
				const ANE_DOUBLE&			refPtY,
				ANE_INT16					numParams,
				const ANE_PTR*		parameters,
				ANE_PTR				myHandle,
				ANE_PTR		reply )
{
	*(ANE_DOUBLE*)reply = sTemp;

} //

//------------------------------------------------------------------------------
// GInitArrayMMFun:
//------------------------------------------------------------------------------
// Function definition:
// 	Define a MeshMaker function "SUTRA_InitArray" 
//
// This function is used by the SUTRA export template
//
//------------------------------------------------------------------------------

static void
GInitArrayMMFun(const ANE_DOUBLE&				refPtX,
				const ANE_DOUBLE&			refPtY,
				ANE_INT16					numParams,
				const ANE_PTR*		parameters,
				ANE_PTR				myHandle,
				ANE_PTR		reply )
{
	ANE_INT32 len = *(ANE_INT32*)(parameters[0]);

	if (sArray)
			delete[] sArray;

	sArray = new double[len];
	sArraySize = len;
	*(ANE_DOUBLE*)reply = 0;
} //


//------------------------------------------------------------------------------
// GArraySetMMFun:
//------------------------------------------------------------------------------
// Function definition:
// 	Define a MeshMaker function "SUTRA_ArraySet" 
//
// This function is used by the SUTRA export template
//
//------------------------------------------------------------------------------

static void
GArraySetMMFun(const ANE_DOUBLE&				refPtX,
				const ANE_DOUBLE&			refPtY,
				ANE_INT16				numParams,
				const ANE_PTR*		parameters,
				ANE_PTR				myHandle,
				ANE_PTR		reply )
{
	ANE_INT32 index = *(ANE_INT32*)(parameters[0]);
	ANE_DOUBLE val = *(ANE_DOUBLE*)(parameters[1]);
	ASSERT(sArray);
	ASSERT(index < sArraySize);
	sArray[index] = val;
	*(ANE_DOUBLE*)reply = 0;
} //

//------------------------------------------------------------------------------
// GArrayGetMMFun:
//------------------------------------------------------------------------------
// Function definition:
// 	Define a MeshMaker function "SUTRA_ArrayGet" 
//
// This function is used by the SUTRA export template
//
//------------------------------------------------------------------------------

static void
GArrayGetMMFun(const ANE_DOUBLE&				refPtX,
				const ANE_DOUBLE&			refPtY,
				ANE_INT16					numParams,
				const ANE_PTR*		parameters,
				ANE_PTR		myHandle,
				ANE_PTR		reply )
{
	ANE_INT32 index = *(ANE_INT32*)(parameters[0]);
	ASSERT(sArray);
	ASSERT(index < sArraySize);
	*(ANE_DOUBLE*)reply = sArray[index];
} //

//------------------------------------------------------------------------------
// GGetDirectoryMMFun:
//------------------------------------------------------------------------------
// Function definition:
//
//	returns the directory of the PIE's DLL.
//
//------------------------------------------------------------------------------

static void
GGetDirectoryMMFun(const ANE_DOUBLE&				refPtX,
				const ANE_DOUBLE&			refPtY,
				ANE_INT16					numParams,
				const ANE_PTR*		parameters,
				ANE_PTR		myHandle,
				ANE_PTR		reply )
{
	char fileName[MAX_PATH];
	char fullFileName[MAX_PATH];
	HINSTANCE hInstance = AfxGetInstanceHandle( );
	GetModuleFileName(hInstance, fullFileName, MAX_PATH);


#ifndef _MAC
	GetShortPathName(fullFileName, fileName, MAX_PATH);
#endif

	int i = strlen(fileName) - 1;
	for (; i > 0 && fileName[i] != '\\' && fileName[i] != '/'  ; i--)
		;
	
	ASSERT(i > 0);
	fileName[i+1] = '\0';

	ANE_STR dirName = (ANE_STR)ANE_MemAlloc(myHandle, strlen(fileName)+1);
	strcpy(dirName, fileName);
	
	*(ANE_STR*)reply = dirName;
} //



//==============================================================================
// SUTRA Template definition
//==============================================================================

//------------------------------------------------------------------------------
// GSutraTemplate:
//------------------------------------------------------------------------------
// Called by MeshMaker when the user ask to Export for SUTRA
// Show project-information dialog for the last time,
// then construct the export template for that project
//------------------------------------------------------------------------------

void GSutraTemplate(ANE_PTR aneHandle, ANE_STR* returnTemplate)
{

	TRY
	{
		
		SUTRAProjectInfo* info;
		ANE_GetPIEProjectHandle(aneHandle,(ANE_PTR*)&info);

		CRunSUTRADialog runSUTRADialog;

		runSUTRADialog.m_ANEHandle = aneHandle;
		runSUTRADialog.m_ProjectInfo = info;

		if (runSUTRADialog.DoModal() == IDOK)
		{
			
			long runSutra = runSUTRADialog.m_RunSutra == 0;
			long alertLevel = runSUTRADialog.m_AlertLevel;
			long calibrate = runSUTRADialog.m_Calibrate;
			//
			// prepare string variables for each of the numerical
			// variables in the SUTRA project information block


//========================================================
// Build export template string
//========================================================

			CString exportTemplate;

			if (!SReadFromFile("SUTRA.MET", exportTemplate))
			{
					exportTemplate = 
"# This is only an error message\n\
#\n\
Alert: \"The MeshMaker Export Template file SUTRA.MET was not found in the same directory as file SUTRA.DLL, please obtain a copy of this file and copy it there.\"\n\
"
;
			}
				
			SReplaceInt(exportTemplate, "@RUNSUTRA@", runSutra);
			SReplaceInt(exportTemplate, "@ALERT@", alertLevel);
			SReplaceInt(exportTemplate, "@Calibrate@", calibrate);
			SReplaceInt(exportTemplate, "@transportConditions@", info->transportConditions);
			
			// Pane1:

			SReplaceInt(exportTemplate, "@NTOBS@", info->NTOBS);
			SReplaceString(exportTemplate, "@SIMULA@", info->SIMULA);
			SReplaceString(exportTemplate, "@TITLE1@", info->TITLE1);
			SReplaceString(exportTemplate, "@TITLE2@", info->TITLE2);

			// Pane2:

			SReplaceDouble(exportTemplate, "@GNUP@", info->GNUP);
			SReplaceDouble(exportTemplate, "@GNUU@", info->GNUU);
			SReplaceInt(exportTemplate, "@IREAD@", info->IREAD);
			SReplaceInt(exportTemplate, "@ISSFLO@", info->ISSFLO);
			SReplaceInt(exportTemplate, "@ISSTRA@", info->ISSTRA);
			SReplaceInt(exportTemplate, "@ISTORE@", info->ISTORE);
			SReplaceInt(exportTemplate, "@IUNSAT@", info->IUNSAT);
			SReplaceDouble(exportTemplate, "@UP@", info->UP);

			// Pane3:

			SReplaceDouble(exportTemplate, "@DELT@", info->DELT);
			SReplaceDouble(exportTemplate, "@DTMAX@", info->DTMAX);
			SReplaceInt(exportTemplate, "@ITCYC@", info->ITCYC);
			SReplaceInt(exportTemplate, "@ITMAX@", info->ITMAX);
			SReplaceInt(exportTemplate, "@NPCYC@", info->NPCYC);
			SReplaceInt(exportTemplate, "@NUCYC@", info->NUCYC);
			SReplaceDouble(exportTemplate, "@DTMULT@", info->DTMULT);
			SReplaceDouble(exportTemplate, "@TMAX@", info->TMAX);

			// Pane4:

			SReplaceInt(exportTemplate, "@KBUDG@", info->KBUDG);
			SReplaceInt(exportTemplate, "@KELMNT@", info->KELMNT);
			SReplaceInt(exportTemplate, "@KINCID@", info->KINCID);
			SReplaceInt(exportTemplate, "@KNODAL@", info->KNODAL);
			SReplaceInt(exportTemplate, "@KPLOTP@", info->KPLOTP);
			SReplaceInt(exportTemplate, "@KPLOTU@", info->KPLOTU);
			SReplaceInt(exportTemplate, "@KVEL@", info->KVEL);
			SReplaceInt(exportTemplate, "@NPRINT@", info->NPRINT);

			// Pane5:

			SReplaceInt(exportTemplate, "@ITRMAX@", info->ITRMAX);
			SReplaceDouble(exportTemplate, "@RPMAX@", info->RPMAX);
			SReplaceDouble(exportTemplate, "@RUMAX@", info->RUMAX);

			// Pane6:

			SReplaceDouble(exportTemplate, "@COMPFL@", info->COMPFL);
			SReplaceDouble(exportTemplate, "@CW@", info->CW);
			SReplaceDouble(exportTemplate, "@DRWDU@", info->DRWDU);
			SReplaceDouble(exportTemplate, "@RHOW0@", info->RHOW0);
			SReplaceDouble(exportTemplate, "@SIGMAW@", info->SIGMAW);
			SReplaceDouble(exportTemplate, "@URHOW0@", info->URHOW0);
			SReplaceDouble(exportTemplate, "@VISCO@", info->VISCO);

			// Pane7:

			SReplaceString(exportTemplate, "@ADSMOD@", info->ADSMOD);
			SReplaceDouble(exportTemplate, "@CHI1@", info->CHI1);
			SReplaceDouble(exportTemplate, "@CHI2@", info->CHI2);
			SReplaceDouble(exportTemplate, "@COMPMA@", info->COMPMA);
			SReplaceDouble(exportTemplate, "@CS@", info->CS);
			SReplaceDouble(exportTemplate, "@RHOS@", info->RHOS);
			SReplaceDouble(exportTemplate, "@SIGMAS@", info->SIGMAS);

			// Pane8:

			SReplaceDouble(exportTemplate, "@GRAVX@", info->GRAVX);
			SReplaceDouble(exportTemplate, "@GRAVY@", info->GRAVY);
			SReplaceDouble(exportTemplate, "@PRODF0@", info->PRODF0);
			SReplaceDouble(exportTemplate, "@PRODF1@", info->PRODF1);
			SReplaceDouble(exportTemplate, "@PRODS0@", info->PRODS0);
			SReplaceDouble(exportTemplate, "@PRODS1@", info->PRODS1);

			// Pane9:

			SReplaceInt(exportTemplate, "@IDIREC@", info->IDIREC);
			SReplaceInt(exportTemplate, "@NCHAPI@", info->NCHAPI);
			SReplaceInt(exportTemplate, "@NCHAPL@", info->NCHAPL);
			SReplaceInt(exportTemplate, "@NLINPI@", info->NLINPI);
			SReplaceDouble(exportTemplate, "@PBASE@", info->PBASE);
			SReplaceDouble(exportTemplate, "@UBASE@", info->UBASE);

			// Pane10:

			SReplaceInt(exportTemplate, "@NOBCYC@", info->NOBCYC);
			SReplaceDouble(exportTemplate, "@TSTART@", info->TSTART);


//========================================================
// End of layer structure definition
//========================================================

			//
			// prepare a characters buffer to return
			// the template to MeshMaker

			char* buf = new char[exportTemplate.GetLength()+1];
			strcpy(buf, exportTemplate);
			
			*returnTemplate = buf;

		}
		else
		{
			//
			// In case the user clicked Cancel,
			// return blank template

			static char empty[] = "";

			*returnTemplate = empty;

		}
	}
	CATCH_ALL(e)
	{
		// return FALSE;
	}
	END_CATCH_ALL
}




//=================================================================
//	FishNet Import PIE functions and declerations
//=================================================================

/*-------------------------------------------------------------------------
// sReserve
//--------------------------------------------------------------------------
// this function is used to handle dynamic array
//--------------------------------------------------------------------------*/
 
static void sReserve(char** buf, long* bufLen, long minLength)
{
	long newLen;

	if (*bufLen > minLength)
		return;

	newLen = (long) (*bufLen*1.5);
	if (newLen <= minLength)
		newLen = minLength + 32;

	if (*buf)
		*buf = (char*)realloc(*buf, newLen);
	else
		*buf = (char*)malloc(newLen);
	*bufLen = newLen;

} // sReserve

static char* sOutputBuffer;
static long sOutputBufferMaxSize;
static long sOutputBufferLen;

static void sInitOutputBuffer()
{
	sOutputBuffer = 0;
	sOutputBufferMaxSize = 0;
	sOutputBufferLen = 0;

	sReserve(&sOutputBuffer, &sOutputBufferMaxSize, 100);
	sOutputBufferLen = 0;
	sOutputBuffer[0] = '\0'; 
}

static void sWriteString(char* str)
{
	sReserve(&sOutputBuffer, &sOutputBufferMaxSize, sOutputBufferLen+strlen(str)+1);
	strcat(sOutputBuffer+sOutputBufferLen, str);
	sOutputBufferLen += strlen(str);
}

static void sWriteInt(long number)
{
	static char numberBuf[100];
	sprintf(numberBuf, "%ld", number);
	sWriteString(numberBuf);
}

static void sWriteDouble(double number)
{
	static char numberBuf[100];
	sprintf(numberBuf, "%lg", number);
	sWriteString(numberBuf);
}

static char* sGetOutputBuffer()
{
	return sOutputBuffer;
}

static void sTerminateOutputBuffer()
{
	free(sOutputBuffer);
}

//==============================================================================
// ImportPIE definitions
//==============================================================================

//
// some struct declerations that we will use

struct SPoint							// 2D point struct
	{
		double x;
		double y;

	}; // SPoint

struct SPatch						// a quadrilateral patch struct
	{
		SPoint point[4];			// 4 points of the patch
		long nodeNumber[4];		// 4 node numbers of the 4 points
		long neighbour[4];			// patch number of neighbour pathces
		int backindex[4];			// number of neighbour pathces
		double density[2];			// x, y densitys
		int densWasCorrected[2];		// true after correcting the density

	}; // SPatch

typedef SPatch* SPatchPtr;

struct SElement					// a quadrilateral element
	{
		long nodeNumber[4];		// 4 node numbers of the 4 points

	}; // SElement


	//
	// definition for patches array

	SPatchPtr*	patches = 0;
	long patchesArraySize = 0;
	long numPatches = 0;

	SPoint* nodes = 0;
	long nodesArraySize = 0;
	long currentNodeNumber = 0;


//------------------------------------------------------------------------------
// MIN/MAX:
//------------------------------------------------------------------------------
// Utility macro to get the min/max value of 2 numbers
//------------------------------------------------------------------------------

#define MIN(a,b) ((a)<(b)?(a):(b))
#define MAX(a,b) ((a)>(b)?(a):(b))


//------------------------------------------------------------------------------
// PointsEqual:
//------------------------------------------------------------------------------
// Utility function to compare two points for equality.
// Usually the two points are not perfectly equal up to the last significant digit,
// so compare up to 6 digits
//------------------------------------------------------------------------------

int PointsEqual(SPoint* pt1, SPoint* pt2)
{
	double xdom = (pt1->x==0 ? 1 : pt1->x);
	double ydom = (pt1->y==0 ? 1 : pt1->y);

	return fabs( (pt1->x - pt2->x) / xdom ) < 1E-6 &&  fabs( (pt1->y - pt2->y) / ydom ) < 1E-6;

} // PointsEqual


//------------------------------------------------------------------------------
// GetDensityIndex:
//------------------------------------------------------------------------------
// figure out if the 'neighbour' is on the x direction (return 0)
// or on the y direction (return 1)
//------------------------------------------------------------------------------

int GetDensityIndex(SPatch* patch, int neighbour)
{
	double maxdy = 0;
	long index;

	for (int i=0; i<4; i++)
	{
		double y1 = patch->point[i].y;
		double x1 = patch->point[i].x;
		double y2 = patch->point[(i+1)%4].y;
		double x2 = patch->point[(i+1)%4].x;

		if (y1 > y2)																		// check that this edge go down
		{
			if (x1-x2==0)																	// if the edge is vertical, all done
			{																							// this edge MUST be in the y direction!
				index = i;																	// save it's index
				break;																			// no need for more iterations
			}
			else if ( fabs((y1-y2)/(x1-x2)) > maxdy)						// look for the steepest edge
			{																							// this edge is in the y direction!
				index = i;																	// save it's index
				maxdy = fabs((y1-y2)/(x1-x2)); //y1-y2;
			}
		}
	} // for i

	return neighbour%2==index%2?1:0;										// if the nieghbour modulo 2 is the same as the found
																								// edge index modulo 2, it's the y direction (return 1)

} // GetDensityIndex


//------------------------------------------------------------------------------
// GetPatchNeighbourIndex:
//------------------------------------------------------------------------------
// Utility function to find the index of 'neighbour' patch within
// the neighbours of 'patch'
//------------------------------------------------------------------------------

int GetPatchNeighbourIndex(SPatch* patch, SPatch* neighbour)
{
	for (int i=0; i<4; i++)
	{
		if (patches[patch->neighbour[i]]==neighbour)
			return i;
	}

	return -1;

} // GetPatchNeighbourIndex


//------------------------------------------------------------------------------
// GetMaxPatchDensity:
//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

long GetMaxPatchDensity(long patchIndex, long neighbourIndex)
{
	if (patchIndex==-1)
		return 0;

	SPatch* patch = patches[patchIndex];
	int densIndex = GetDensityIndex(patch, neighbourIndex);

	// if patch->densWasCorrected[densIndex] is TRUE,
	// that means we are back in the first patch...
	// to prevent infinite loop just return.

	if (patch->densWasCorrected[densIndex])
		return 0;

	int newIndex = (neighbourIndex+2)%4;
	long d = GetMaxPatchDensity(patch->neighbour[newIndex], patch->backindex[newIndex]);
	
	return (long) MAX(d, patch->density[densIndex]);

} // GetMaxPatchDensity


//------------------------------------------------------------------------------
// SetPatchDensity:
//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void SetPatchDensity(long patchIndex, long neighbourIndex, double dens)
{
	if (patchIndex==-1)
		return;

	SPatch* patch = patches[patchIndex];
	int densIndex = GetDensityIndex(patch, neighbourIndex);
	
	//
	// should check for plonters here!@#$%:
	// if (patch->densWasCorrected[densIndex]) { do something, quick! }

	if (patch->densWasCorrected[densIndex])
	{
		// if patch->densWasCorrected[densIndex] if TRUE,
		// that means we're back on the first patch...
		// to prevent infinite loop, just return

		//patch->densWasCorrected[densIndex] = FALSE;
		return;
	}

	int newIndex = (neighbourIndex+2)%4;
	SetPatchDensity(patch->neighbour[newIndex], patch->backindex[newIndex], dens);

	patch->density[densIndex] = dens;
	patch->densWasCorrected[densIndex] = TRUE;

} // SetPatchDensity


//------------------------------------------------------------------------------
// SearchNode:
//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

long SearchNode(SPoint pt, double xepsilon, double yepsilon)
{
	for (long i=0; i<currentNodeNumber; i++)
	{
		if ( fabs(pt.x-nodes[i].x)<=xepsilon && fabs(pt.y-nodes[i].y)<=yepsilon )
			return i;
	}

	return -1;

} // SearchNode


//------------------------------------------------------------------------------
// GPatchedMeshImportProc:
//------------------------------------------------------------------------------
// Import and mesh:
// 	Import a set of (hopefuly) rectangular patches from the source information layer
//	and build a quad-mesh
//------------------------------------------------------------------------------

static void
GPatchedMeshImportProc(ANE_PTR aneHandle, const ANE_STR fileName, ANE_PTR layerHandle)
{
	//
	// prepare array of patches

	patches = 0;
	patchesArraySize = 0;
	numPatches = 0;
	sReserve((char**)&patches, &patchesArraySize, sizeof(SPatch*)*10);			// reserve room for at least 10 patches

	char* ibuff;
	char* obuff;

	//
	// get a string from the source layer describing the patches.
	// the string is in the information layer export form:
	//
	// <num-of-points> <parameter1> <parameter2>
	// <x1> <y1> 
	// <x2> <y2> 
	// <x3> <y3> 
	// <x4> <y4> 
	// <x5> <y5> 
	//
	// Notes:
	//	<num-of-points> should be 5
	//	<parameter1> will be used as the density in the x direction
	//	<parameter2> will be used as the density in the y direction
	//	<x5> and <y5> should be the same as <x1> and <y1>!

	layerHandle = ANE_LayerGetHandleByName(aneHandle, "FishNet Mesh Layout");
	if (layerHandle==0)
	{
		ANE_UserMessageOk(aneHandle, "Layer: 'FishNet Mesh Layout' not found!");
		return;
	}

	ANE_ExportTextFromOtherLayer(aneHandle, layerHandle, &ibuff);

	//
	// prepare to read the patches

	istrstream importText(ibuff);
	long numpoints;
	double x, y;
	int problemEncountered = FALSE;
	SPatch* patch;

	//
	// loop and read all the patches

	while(TRUE)
	{
		//assert(0);
		//
		// skip comments and empty lines

		/*do		
		{
//			importText.ignore(256, '\n');
			importText.ignore(256, '\r');
		} */
		while (importText.peek() == '#' || importText.peek() == '\n'||//;
		importText.peek() == '\r')
			importText.ignore(256, '\n');

		if (!importText || importText.eof())
			break;

		//
		// read header line

		importText >> numpoints;					// should be 5 points
		if (numpoints != 5 || importText.eof())
		{
			problemEncountered = TRUE;
			break;
		}

		patch = new SPatch;
		importText >> patch->density[0];		if (importText.eof()) break;
		importText >> patch->density[1];

		patch->density[0]--;				// decrement count, to get number of lines
		patch->density[1]--;				// out of row/columns count

		patch->densWasCorrected[0] = FALSE;
		patch->densWasCorrected[1] = FALSE;

		//
		// skip comment and empty lines

		/*do 
		{
//			importText.ignore(256, '\n');
			importText.ignore(256, '\r');
		} */
//		assert(0);
		while (importText.peek() == '#' || importText.peek() == '\n'||//;
			importText.peek() == '\r')
			importText.ignore(256, '\n');

		if (!importText || importText.eof())
			break;

		//
		// read x and y values of the 5 points

		importText >> patch->point[0].x;
		importText >> patch->point[0].y;
		patch->nodeNumber[0] = -1;				// reset node number
		patch->neighbour[0] = -1;					// reset neighbour number

		importText >> patch->point[1].x;
		importText >> patch->point[1].y;
		patch->nodeNumber[1] = -1;				// reset node number
		patch->neighbour[1] = -1;					// reset neighbour number

		importText >> patch->point[2].x;
		importText >> patch->point[2].y;
		patch->nodeNumber[2] = -1;				// reset node number
		patch->neighbour[2] = -1;					// reset neighbour number

		importText >> patch->point[3].x;
		importText >> patch->point[3].y;
		patch->nodeNumber[3] = -1;				// reset node number
		patch->neighbour[3] = -1;					// reset neighbour number

		importText >> x;								// should be equal to patch->point[0].x
		importText >> y;								// should be equal to patch->point[0].y

		if (!importText)								// check for a problem
			break;

		//
		// add the new patch to the list

		sReserve((char**)&patches, &patchesArraySize, sizeof(SPatch*)*(numPatches+1));		// reserve room for at least 'numPatches' patches
		patches[numPatches++] = patch;																						// insert patch into array
	};

//???	ANE_MemDelete(aneHandle, ibuff);							// clear the block of memory returned from 'ANE_ExportTextFromOtherLayer()'

	//
	// prepare an array of node positions
	// max possible nodes are 4 times the patches count

	nodes = 0;
	nodesArraySize = 0;
	currentNodeNumber = 0;
	sReserve((char**)&nodes, &nodesArraySize, sizeof(SPoint)*(numPatches*4+1));										// reserve room for nodes

	//
	// prepare an array of elements

	SElement* elements = 0;
	long numElements = 0;
	long elemsArraySize = 0;
	sReserve((char**)&elements, &elemsArraySize, sizeof(SElement)*(numPatches*4+1));				// reserve room for elements

	//

	long i, j, k, l;
	// loop for all patches and calc the node numbers:

	for (i=0; i<numPatches; i++)																									// for all patches
	{
		patch = patches[i];																											// save a pointer to the current patch
		
		//
		// loop for all nodes and assign them the proper node-numbers

		for (j=0; j<4; j++)																											// loop for patch's 9 nodes
		{
			//
			// search all known nodes, is it equal to our vertex?

			int nodeFound = FALSE;
			for (k=0; k<currentNodeNumber && !nodeFound; k++)													// check all known nodes
				if (PointsEqual(&nodes[k], &patch->point[j]))															// test equality
					nodeFound = TRUE;																								// a node was found
			
			if (nodeFound)
			{
				// a node was found
				// take it's number
				
				patch->nodeNumber[j] = k-1;																					// take the found node number
			}
			else
			{
				// a node was NOT found
				// create a new node and
				// take a new number
				
				patch->nodeNumber[j] = currentNodeNumber;															// take the found node number
				nodes[currentNodeNumber++] = patch->point[j];														// save the node's position
			}

		} // for j

	} // for i

	//
	// figure out patch neighbours

	for (i=0; i<numPatches; i++)																									// for all patches
	{
		patch = patches[i];																											// save a pointer to the current patch
		
		//
		// loop for all nodes and assign them the proper node-numbers

		for (j=0; j<4; j++)																											// loop for patch's 4 nodes
			if (patch->neighbour[j]==-1)
			{
				long n1 = patch->nodeNumber[j];
				long n2 = patch->nodeNumber[(j+1)%4];

				// search neighbour patch
				
				int found = FALSE;
				for (k=0; k<numPatches && !found; k++)
				{
					SPatch* p =  patches[k];
					
					for (l=0; l<4 && !found; l++)
						if (p->nodeNumber[l]==n2 && p->nodeNumber[(l+1)%4]==n1 )
						{
							found = TRUE;
							patch->neighbour[j]=k;
							patch->backindex[j]=l;

							p->neighbour[l]=i;
							p->backindex[l]=j;
						} // for l, if neighbour found

				} // for k

			} // for j

	} // for i

	//
	// prepare meshing densities

	for (i=0; i<numPatches; i++)																									// for all patches
	{
		long dens1, dens2, dens, densIndex;

		patch = patches[i];																											// save a pointer to the current patch

		densIndex = GetDensityIndex(patch, 0);

		if (!patch->densWasCorrected[densIndex])
		{
			patch->densWasCorrected[densIndex] = TRUE;

			dens1 = GetMaxPatchDensity(patch->neighbour[0], patch->backindex[0] );
			dens2 = GetMaxPatchDensity(patch->neighbour[2], patch->backindex[2] );

			dens  = (long) MAX(patch->density[densIndex], MAX(dens1, dens2));

			patch->density[densIndex] = dens;
//			patch->densWasCorrected[densIndex] = TRUE;

			SetPatchDensity(patch->neighbour[0], patch->backindex[0], dens);
			SetPatchDensity(patch->neighbour[2], patch->backindex[2], dens);
		
		}

		densIndex = (densIndex+1)%2;

		if (!patch->densWasCorrected[densIndex])
		{
			patch->densWasCorrected[densIndex] = TRUE;

			dens1 = GetMaxPatchDensity(patch->neighbour[1], patch->backindex[1] );
			dens2 = GetMaxPatchDensity(patch->neighbour[3], patch->backindex[3] );

			dens  = (long) MAX(patch->density[densIndex], MAX(dens1, dens2));

			patch->density[densIndex] = dens;
//			patch->densWasCorrected[densIndex] = TRUE;

			SetPatchDensity(patch->neighbour[1], patch->backindex[1], dens);
			SetPatchDensity(patch->neighbour[3], patch->backindex[3], dens);
		}

	} // for i


	//**************
	// do the meshing here
	//**************


	for (i=0; i<numPatches; i++)																									// for all patches
	{
		patch = patches[i];																											// save a pointer to the current patch

		double maxx = MAX(patch->point[0].x, MAX(patch->point[1].x, MAX(patch->point[2].x, patch->point[3].x)));
		double minx = MIN(patch->point[0].x, MIN(patch->point[1].x, MIN(patch->point[2].x, patch->point[3].x)));
		double maxy = MAX(patch->point[0].y, MAX(patch->point[1].y, MAX(patch->point[2].y, patch->point[3].y)));
		double miny = MIN(patch->point[0].y, MIN(patch->point[1].y, MIN(patch->point[2].y, patch->point[3].y)));
		double xepsilon = 1e-6*(maxx-minx);
		double yepsilon = 1e-6*(maxy-miny);


		long dens1, dens2, densIndex;
		densIndex = GetDensityIndex(patch, 0);
		dens1 = (long) patch->density[densIndex];
		dens2 = (long) patch->density[(densIndex+1)%2];

		long numNodes = (dens1+2)*(dens2+2);
		long* nodeNumber = 0;
		long nodeNumberArraySize = 0;
		sReserve((char**)&nodeNumber, &nodeNumberArraySize, sizeof(long)*(numNodes+1));		// reserve room for node numbers

		for (j=0; j<numNodes; j++)
			nodeNumber[j] = -1;

		//
		// save patch's 4 nodes

		nodeNumber[0]										= patch->nodeNumber[0];
		nodeNumber[dens1+1]							= patch->nodeNumber[1];
		nodeNumber[(dens1+2)*(dens2+2)-1]	= patch->nodeNumber[2];
		nodeNumber[(dens1+2)*(dens2+1)]		= patch->nodeNumber[3];

		for (j=0; j<dens2+2; j++)
			for (k=0; k<dens1+2; k++)
			{
				long n = j*(dens1+2)+k;
				if (nodeNumber[n] == -1)
				{
					sReserve((char**)&nodes, &nodesArraySize, sizeof(SPoint)*(currentNodeNumber+2));										// reserve room for nodes

					SPoint p1, p2, pt;
					double ratio;
					
					// pt1 is between node[0] and node[3]
					
					ratio = (double)(j)/(dens2+1);
					p1.x = patch->point[0].x*(1-ratio) + patch->point[3].x*ratio;
					p1.y = patch->point[0].y*(1-ratio) + patch->point[3].y*ratio;
	
					// pt2 is between node[1] and node[2]
					
					p2.x = patch->point[1].x*(1-ratio) + patch->point[2].x*ratio;
					p2.y = patch->point[1].y*(1-ratio) + patch->point[2].y*ratio;
					
					// our point should be between p1 and p2

					ratio = (double)(k)/(dens1+1);
					pt.x = p1.x*(1-ratio) + p2.x*ratio;
					pt.y = p1.y*(1-ratio) + p2.y*ratio;
					
					long foundNode = SearchNode(pt, xepsilon, yepsilon);
					if (foundNode == -1)
					{
						nodeNumber[n] = currentNodeNumber;
						nodes[currentNodeNumber++] = pt;
					}
					else
					{
						nodeNumber[n] = foundNode;
					}
				}
			}

		//
		// reserve space for new elements

		long newElements = (dens1+1) * (dens2+1);
		sReserve((char**)&elements, &elemsArraySize, sizeof(SElement)*(numElements+newElements+1));		// reserve room for elements

		//
		// fill new elements with the proper node numbers
		
		for (j=0; j<dens2+1; j++)
			for (k=0; k<dens1+1; k++)
			{
				SElement element;

				element.nodeNumber[0] = nodeNumber[j*(dens1+2)+k];
				element.nodeNumber[1] = nodeNumber[j*(dens1+2)+k+1];
				element.nodeNumber[2] = nodeNumber[(j+1)*(dens1+2)+k+1];
				element.nodeNumber[3] = nodeNumber[(j+1)*(dens1+2)+k];
	
				elements[numElements++] = element;																											// insert patch into array
			}

		delete nodeNumber;								// clear the block of memory used for the node-numbers

	} // for i

	delete patches;										// clear the block of memory used for the patches

	//
	// prepare output string
	// the string should be in the mesh layer's import form:
	//
	// <num-of-elements> <num-of-nodes> <num-of-element-parameters> <num-of-node-parameters>
	// N <node-number1> <x1> <y1> 
	// N <node-number2> <x2> <y2> 
	// .
	// .
	// E <elem-number1> <node1> <node2> <node3> <node4>
	// E <elem-number2> <node1> <node2> <node3> <node4>
	// .
	// .
	//
	// Notes:
	//	<num-of-element-parameters> and <num-of-node-parameters>
	//	 are both 0 in our example

	sInitOutputBuffer();
//	ostrstream exportText;
	
	// export header line

	sWriteInt(numElements);
	sWriteString(" ");
	sWriteInt(currentNodeNumber);
	sWriteString(" 0 0\n");
//	exportText << numElements << " " << currentNodeNumber << " 0 0\n";

	// export node lines

	for (i=0; i<currentNodeNumber; i++)
	{
		sWriteString("N ");
		sWriteInt(i+1);
		sWriteString(" ");
		sWriteDouble(nodes[i].x);
		sWriteString(" ");
		sWriteDouble(nodes[i].y);
		sWriteString("\n");
		// exportText << "N " << i+1 << " " << nodes[i].x << " " << nodes[i].y << "\n";
	}

	// export element lines

	for (i=0; i<numElements; i++)
	{
		sWriteString("E ");
		sWriteInt(i+1);
		sWriteString(" ");
		sWriteInt(elements[i].nodeNumber[0]+1);
		sWriteString(" ");
		sWriteInt(elements[i].nodeNumber[1]+1);
		sWriteString(" ");
		sWriteInt(elements[i].nodeNumber[2]+1);
		sWriteString(" ");
		sWriteInt(elements[i].nodeNumber[3]+1);
		sWriteString("\n");
//		exportText << "E " << i+1 << " " << elements[i].nodeNumber[0]+1
//												 << " " << elements[i].nodeNumber[1]+1
//												 << " " << elements[i].nodeNumber[2]+1
//												 << " " << elements[i].nodeNumber[3]+1 << "\n";
	}

	obuff = sGetOutputBuffer();

	ANE_ImportTextToLayer(aneHandle, obuff);

	sTerminateOutputBuffer();
	//
	// cleanup
	
	delete nodes;									// clear the block of memory used for the nodes
	delete elements;								// clear the block of memory used for the elements
	return;

} // GPatchedMeshImportProc


//=================================================================
//	PIE handling functions and declerations
//=================================================================

struct ANEPIEDesc* gFunDesc[40];							// list of PIE descriptors for all parts


//
// "SUTRA_Run" function part

struct ANEPIEDesc				gRunSUTRADesc;				// PIE defcriptor
struct FunctionPIEDesc			gRunSUTRAFDesc;				// Function descriptor
char* gpnNone[1] = {0};										// list of (empty) parameters names list for function
enum EPIENumberType	gNoParamTypes[1] = {(EPIENumberType)0};	// list of (empty) parameters types list for function

//
// "Run SUTRA" export template part

struct ANEPIEDesc				gSutraTemplateDesc;			// PIE defcriptor
struct ExportTemplatePIEDesc	gSutraTemplateTDesc;		// Export template descriptor

//
// "SUTRA" project part

struct ANEPIEDesc 				gSUTRAProjectPIEDesc;		// PIE defcriptor
struct ProjectPIEDesc			gSUTRAProjectPPIEDesc;		// Project descriptor



//
// "SUTRA_xx" functions part

struct ANEPIEDesc				gSetTempDesc;				// PIE defcriptor
struct FunctionPIEDesc			gSetTempFDesc;				// Function descriptor
char* gpnValue[2] = {"Value", 0};							// list of (empty) parameters names list for function
enum EPIENumberType	gOneDoubleTypes[2] = {kPIEFloat, (EPIENumberType)0};	// list of (empty) parameters types list for function

struct ANEPIEDesc				gGetTempDesc;				// PIE defcriptor
struct FunctionPIEDesc			gGetTempFDesc;				// Function descriptor

struct ANEPIEDesc				gInitArrayDesc;				// ANEPIE defcriptor
struct FunctionPIEDesc			gInitArrayFDesc;			// Function descriptor
char* gpnLength[2] = {"Length", 0};										// list of (empty) parameters names list for function
enum EPIENumberType	gOneIntegerTypes[2] = {kPIEInteger, (EPIENumberType)0};	// list of (empty) parameters types list for function

struct ANEPIEDesc				gArraySetDesc;				// PIE defcriptor
struct FunctionPIEDesc			gArraySetFDesc;				// Function descriptor
char* gpnIndexValue[3] = {"Index", "Value", 0};										// list of (empty) parameters names list for function
enum EPIENumberType	gIntegerDoubleTypes[3] = {kPIEInteger,kPIEFloat, (EPIENumberType)0};	// list of (empty) parameters types list for function

struct ANEPIEDesc				gArrayGetDesc;				// PIE defcriptor
struct FunctionPIEDesc			gArrayGetFDesc;				// Function descriptor
char* gpnIndex[2] = {"Index", 0};										// list of (empty) parameters names list for function

struct ANEPIEDesc				gGetDirDesc;				// PIE Descriptor
struct FunctionPIEDesc			gGetDirFDesc;				// PIE Descriptor

struct ANEPIEDesc				gFishNetDesc;				// PIE Descriptor
struct ImportPIEDesc			gFishNetIDesc;				// PIE Descriptor

//
// "SUTRA Post Processing Import"  part

struct ANEPIEDesc 				gSUTRAPPImportPIEDesc;		// PIE defcriptor
struct ImportPIEDesc			gSUTRAPPPIPIEDesc;			// ImportPIE descriptor

extern "C" {
void
GetANEFunctions(ANE_INT32& numNames, struct ANEPIEDesc**& descriptors);
}

void
GetANEFunctions(ANE_INT32& numNames, struct ANEPIEDesc**& descriptors)
{ 
	numNames = 0;

	//
	// prepare FUNCTION descriptor for "SUTRA_Run" function

	gRunSUTRAFDesc.version = FUNCTION_PIE_VERSION;		// set correct version
	gRunSUTRAFDesc.name = "SUTRA_Run";							// name of function
	gRunSUTRAFDesc.address = GRunSUTRAMMFun;					// function address
	gRunSUTRAFDesc.functionFlags = (EFunctionPIEFlags)(kFunctionIsHidden | kFunctionNeedsProject);		// set correct version
	gRunSUTRAFDesc.neededProject = "SUTRA";					// available only for SUTRA projects
	gRunSUTRAFDesc.returnType = kPIEFloat;						// return value type
	gRunSUTRAFDesc.numParams =  0;							// number of parameters
	gRunSUTRAFDesc.numOptParams = 0;							// number of optional parameters
	gRunSUTRAFDesc.paramNames = gpnNone;						// pointer to parameter names list
	gRunSUTRAFDesc.paramTypes = gNoParamTypes;				// pointer to parameters types list
	
	//
	// prepare PIE descriptor for "SUTRA_RunSUTRA" function

	gRunSUTRADesc.name  = "SUTRA_Run";							// name of PIE
	gRunSUTRADesc.type =  kFunctionPIE;						// PIE type: PIE function
	gRunSUTRADesc.descriptor = &gRunSUTRAFDesc;					// pointer to descriptor
	
	gFunDesc[(numNames)++] = &gRunSUTRADesc;					// add descriptor to list

	//
	// prepare TEMPLATE descriptor for "Run SUTRA" export template

	gSutraTemplateTDesc.name = "Run &SUTRA";						// name for export menu
	gSutraTemplateTDesc.exportType =  kPIEQuadMeshLayer;			// type of template
	gSutraTemplateTDesc.exportFlags = (EExportTemplateFlags)( kExportDontShowParamDialog | kExportNeedsProject) /* | kExportDontShowFileDialog */;
	gSutraTemplateTDesc.getTemplateProc = GSutraTemplate;		// function address
	gSutraTemplateTDesc.neededProject = "SUTRA";		// function address

	//
	// prepare PIE descriptor for "Run SUTRA" export template

	gSutraTemplateDesc.name = "Run &SUTRA";						// PIE name
	gSutraTemplateDesc.type = kExportTemplatePIE;			// PIE type: export template
	gSutraTemplateDesc.descriptor = &gSutraTemplateTDesc;		// pointer to descriptor

	gFunDesc[(numNames)++] = &gSutraTemplateDesc;			// add descriptor to list

	//
	// prepare PROJECT descriptor for SUTRA project
	
	gSUTRAProjectPPIEDesc.name = "SUTRA";						// name of project
	gSUTRAProjectPPIEDesc.projectFlags = (EProjectPIEFlags)(kProjectShouldSave | kProjectCanEdit | kProjectShouldClean |kProjectDisplaysDialog);
	gSUTRAProjectPPIEDesc.createNewProc = GSUTRAProjectNew;		// address of 'New' function 
	gSUTRAProjectPPIEDesc.editProjectProc = GSUTRAProjectEdit;	// address of 'Edit' function
	gSUTRAProjectPPIEDesc.cleanProjectProc = GSUTRAProjectClean;// address of 'Clean' function
	gSUTRAProjectPPIEDesc.saveProc = GSUTRAProjectSave;			// address of 'Save' function
	gSUTRAProjectPPIEDesc.loadProc = GSUTRAProjectLoad;			// address of 'Load' function

	//
	// prepare PIE descriptor for "SUTRA" project

	gSUTRAProjectPIEDesc.name = "New SUTRA Project...";			// PIE name
	gSUTRAProjectPIEDesc.type = kProjectPIE;					// PIE type: project PIE
	gSUTRAProjectPIEDesc.descriptor = &gSUTRAProjectPPIEDesc;	// pointer to descriptor
	
	gFunDesc[(numNames)++] = &gSUTRAProjectPIEDesc;			// add descriptor to list



	//
	// prepare FUNCTION descriptor for "SUTRA_SetTemp" function

	gSetTempFDesc.version = FUNCTION_PIE_VERSION;		// set correct version
	gSetTempFDesc.name = "SUTRA_SetTemp";							// name of function
	gSetTempFDesc.address = GSetTempMMFun;					// function address
	gSetTempFDesc.functionFlags = (EFunctionPIEFlags)(kFunctionIsHidden | kFunctionNeedsProject);		// set correct version
	gSetTempFDesc.neededProject = "SUTRA";					// available only for SUTRA projects
	gSetTempFDesc.returnType = kPIEFloat;						// return value type
	gSetTempFDesc.numParams =  1;							// number of parameters
	gSetTempFDesc.numOptParams = 0;							// number of optional parameters
	gSetTempFDesc.paramNames = gpnValue;						// pointer to parameter names list
	gSetTempFDesc.paramTypes = gOneDoubleTypes;				// pointer to parameters types list
	
	//
	// prepare PIE descriptor for "SUTRA_SetTemp" function

	gSetTempDesc.name  = "SUTRA_SetTemp";							// name of PIE
	gSetTempDesc.type =  kFunctionPIE;						// PIE type: PIE function
	gSetTempDesc.descriptor = &gSetTempFDesc;					// pointer to descriptor
	
	gFunDesc[(numNames)++] = &gSetTempDesc;					// add descriptor to list



	//
	// prepare FUNCTION descriptor for "SUTRA_GetTemp" function

	gGetTempFDesc.version = FUNCTION_PIE_VERSION;		// set correct version
	gGetTempFDesc.name = "SUTRA_GetTemp";							// name of function
	gGetTempFDesc.address = GGetTempMMFun;					// function address
	gGetTempFDesc.functionFlags =(EFunctionPIEFlags)(kFunctionIsHidden | kFunctionNeedsProject);		// set correct version
	gGetTempFDesc.neededProject = "SUTRA";					// available only for SUTRA projects
	gGetTempFDesc.returnType = kPIEFloat;						// return value type
	gGetTempFDesc.numParams =  0;							// number of parameters
	gGetTempFDesc.numOptParams = 0;							// number of optional parameters
	gGetTempFDesc.paramNames = gpnNone;						// pointer to parameter names list
	gGetTempFDesc.paramTypes = gNoParamTypes;				// pointer to parameters types list
	
	//
	// prepare PIE descriptor for "SUTRA_GetTemp" function

	gGetTempDesc.name  = "SUTRA_GetTemp";							// name of PIE
	gGetTempDesc.type =  kFunctionPIE;						// PIE type: PIE function
	gGetTempDesc.descriptor = &gGetTempFDesc;					// pointer to descriptor
	
	gFunDesc[(numNames)++] = &gGetTempDesc;					// add descriptor to list


	//
	// prepare FUNCTION descriptor for "SUTRA_InitArray" function

	gInitArrayFDesc.version = FUNCTION_PIE_VERSION;		// set correct version
	gInitArrayFDesc.name = "SUTRA_InitArray";							// name of function
	gInitArrayFDesc.address = GInitArrayMMFun;					// function address
	gInitArrayFDesc.functionFlags = (EFunctionPIEFlags)(kFunctionIsHidden | kFunctionNeedsProject);		// set correct version
	gInitArrayFDesc.neededProject = "SUTRA";					// available only for SUTRA projects
	gInitArrayFDesc.returnType = kPIEFloat;						// return value type
	gInitArrayFDesc.numParams =  1;							// number of parameters
	gInitArrayFDesc.numOptParams = 0;							// number of optional parameters
	gInitArrayFDesc.paramNames = gpnLength;						// pointer to parameter names list
	gInitArrayFDesc.paramTypes = gOneIntegerTypes;				// pointer to parameters types list
	
	//
	// prepare PIE descriptor for "SUTRA_InitArray" function

	gInitArrayDesc.name  = "SUTRA_InitArray";							// name of PIE
	gInitArrayDesc.type =  kFunctionPIE;						// PIE type: PIE function
	gInitArrayDesc.descriptor = &gInitArrayFDesc;					// pointer to descriptor
	
	gFunDesc[(numNames)++] = &gInitArrayDesc;					// add descriptor to list

	//
	// prepare FUNCTION descriptor for "SUTRA_ArrayGet" function

	gArrayGetFDesc.version = FUNCTION_PIE_VERSION;		// set correct version
	gArrayGetFDesc.name = "SUTRA_ArrayGet";							// name of function
	gArrayGetFDesc.address = GArrayGetMMFun;					// function address
	gArrayGetFDesc.functionFlags = (EFunctionPIEFlags)(kFunctionIsHidden | kFunctionNeedsProject);		// set correct version
	gArrayGetFDesc.neededProject = "SUTRA";					// available only for SUTRA projects
	gArrayGetFDesc.returnType = kPIEFloat;						// return value type
	gArrayGetFDesc.numParams =  1;							// number of parameters
	gArrayGetFDesc.numOptParams = 0;							// number of optional parameters
	gArrayGetFDesc.paramNames = gpnIndex;						// pointer to parameter names list
	gArrayGetFDesc.paramTypes = gOneIntegerTypes;				// pointer to parameters types list
	
	//
	// prepare PIE descriptor for "SUTRA_ArrayGet" function

	gArrayGetDesc.name  = "SUTRA_ArrayGet";							// name of PIE
	gArrayGetDesc.type =  kFunctionPIE;						// PIE type: PIE function
	gArrayGetDesc.descriptor = &gArrayGetFDesc;					// pointer to descriptor
	
	gFunDesc[(numNames)++] = &gArrayGetDesc;					// add descriptor to list

	//
	// prepare FUNCTION descriptor for "SUTRA_ArraySet" function

	gArraySetFDesc.version = FUNCTION_PIE_VERSION;		// set correct version
	gArraySetFDesc.name = "SUTRA_ArraySet";							// name of function
	gArraySetFDesc.address = GArraySetMMFun;					// function address
	gArraySetFDesc.functionFlags = (EFunctionPIEFlags)(kFunctionIsHidden | kFunctionNeedsProject);		// set correct version
	gArraySetFDesc.neededProject = "SUTRA";					// available only for SUTRA projects
	gArraySetFDesc.returnType = kPIEFloat;						// return value type
	gArraySetFDesc.numParams =  2;							// number of parameters
	gArraySetFDesc.numOptParams = 0;							// number of optional parameters
	gArraySetFDesc.paramNames = gpnIndexValue;						// pointer to parameter names list
	gArraySetFDesc.paramTypes = gIntegerDoubleTypes;				// pointer to parameters types list
	
	//
	// prepare PIE descriptor for "SUTRA_ArrayGet" function

	gArraySetDesc.name  = "SUTRA_ArraySet";							// name of PIE
	gArraySetDesc.type =  kFunctionPIE;						// PIE type: PIE function
	gArraySetDesc.descriptor = &gArraySetFDesc;					// pointer to descriptor
	
	gFunDesc[(numNames)++] = &gArraySetDesc;					// add descriptor to list


	//
	// prepare FUNCTION descriptor for "SUTRA_GetDirectory" function

	gGetDirFDesc.version = FUNCTION_PIE_VERSION;		// set correct version
	gGetDirFDesc.name = "SUTRA_GetDirectory";							// name of function
	gGetDirFDesc.address = GGetDirectoryMMFun;					// function address
	gGetDirFDesc.functionFlags =(EFunctionPIEFlags)(kFunctionIsHidden | kFunctionNeedsProject);		// set correct version
	gGetDirFDesc.neededProject = "SUTRA";					// available only for SUTRA projects
	gGetDirFDesc.returnType = kPIEString;						// return value type
	gGetDirFDesc.numParams =  0;							// number of parameters
	gGetDirFDesc.numOptParams = 0;							// number of optional parameters
	gGetDirFDesc.paramNames = gpnNone;						// pointer to parameter names list
	gGetDirFDesc.paramTypes = gNoParamTypes;				// pointer to parameters types list

	//
	// prepare PIE descriptor for "SUTRA_GetDirectory" function

	gGetDirDesc.name  = "SUTRA_GetDirectory";							// name of PIE
	gGetDirDesc.type =  kFunctionPIE;						// PIE type: PIE function
	gGetDirDesc.descriptor = &gGetDirFDesc;					// pointer to descriptor

	gFunDesc[(numNames)++] = &gGetDirDesc;					// add descriptor to list
	

	//
	// fill in the ImportPIE additional information record

	gFishNetIDesc.version			= IMPORT_PIE_VERSION;		// descriptor version, use constant here...
	gFishNetIDesc.name				= "FishNet Mesh...";		// title of menu
	gFishNetIDesc.importFlags		= (enum EImportPIEFlags)0;	// import from another layer
	gFishNetIDesc.toLayerTypes		= kPIEQuadMeshLayer;		// import only TO quad-mesh layers
	gFishNetIDesc.doImportProc		= GPatchedMeshImportProc;	// the address of the import function

	//
	// fill in the PIE descriptor

	gFishNetDesc.name		= "FishNet";										// PIE name
	gFishNetDesc.type		= kImportPIE;											// type of PIE: ImportPIE
	gFishNetDesc.descriptor	= &gFishNetIDesc;									// address of additional information record

	gFunDesc[(numNames)++] = &gFishNetDesc;					// add descriptor to list

	//
	//
	// prepare Import (Post Processing) descriptor for SUTRA project
	
	gSUTRAPPPIPIEDesc.version = IMPORT_PIE_VERSION;
	gSUTRAPPPIPIEDesc.name = "Sutra Post Proc...";						// name of project
	gSUTRAPPPIPIEDesc.importFlags = (EImportPIEFlags)kImportAllwaysVisible, 
		(EImportPIEFlags)kImportNeedsProject;
	gSUTRAPPPIPIEDesc.toLayerTypes = kPIEAnyLayer/* was kPIETriMeshLayer*/;		
	// gSUTRAPPPIPIEDesc.fromLayerTypes = ;	//not used
	gSUTRAPPPIPIEDesc.doImportProc = GSUTRAPostProc;// address of Post Processing Function function
	// gSUTRAPPPIPIEDesc.neededProject = ;			// not used

	//
	// prepare PIE descriptor for "SUTRA Post Proc" Import

	gSUTRAPPImportPIEDesc.name = "Sutra Post Proc...";			// PIE name
	gSUTRAPPImportPIEDesc.type = kImportPIE;					// PIE type: project PIE
	gSUTRAPPImportPIEDesc.descriptor = &gSUTRAPPPIPIEDesc;	// pointer to descriptor
	
	gFunDesc[(numNames)++] = &gSUTRAPPImportPIEDesc;			// add descriptor to list




	descriptors = gFunDesc;
}


// Logic from SUTRA 1984, Clifford I. Voss
// to automatically calculate SUTRA input parameter NTOBS
void SUTRAProjectInfo::CalcAfterDialog()
{

	double TS = TSTART;
	long JT = 0;

	double tempDELT = DELT;

	do
	{
		JT++;

		if (JT > 1 && (JT % ITCYC == 0))
			tempDELT *= DTMULT;

		if (tempDELT > DTMAX)
			tempDELT = DTMAX;

		TS += tempDELT;

	} while (JT < ITMAX && TS < TMAX);

	if (NOBCYC > 0)
		NTOBS = JT/NOBCYC + 2;
	else
		NTOBS = 0;
}


BOOL SUTRAProjectInfo::ShowDialog()
{
		this->PrepareValString();


		CTabbedDialog tabbedDialog("SUTRA Non-Spatial Information (UNIT K1)");

		CPane1 pane1; tabbedDialog.AddPage(&pane1);
		CPane2 pane2; tabbedDialog.AddPage(&pane2);
		CPane3 pane3; tabbedDialog.AddPage(&pane3);
		CPane4 pane4; tabbedDialog.AddPage(&pane4);
		CPane5 pane5; tabbedDialog.AddPage(&pane5);
		CPane6 pane6; tabbedDialog.AddPage(&pane6);
		CPane7 pane7; tabbedDialog.AddPage(&pane7);
		CPane8 pane8; tabbedDialog.AddPage(&pane8);

		pane1.GetValues(*this);
		pane2.GetValues(*this);
		pane3.GetValues(*this);
		pane4.GetValues(*this);
		pane5.GetValues(*this);
		pane6.GetValues(*this);
		pane7.GetValues(*this);
		pane8.GetValues(*this);

		if (tabbedDialog.DoModal() == IDOK)
		{
			//
			// get user values from the dialog

			pane1.ReturnValues(*this);
			pane2.ReturnValues(*this);
			pane3.ReturnValues(*this);
			pane4.ReturnValues(*this);
			pane5.ReturnValues(*this);
			pane6.ReturnValues(*this);
			pane7.ReturnValues(*this);
			pane8.ReturnValues(*this);

			this->CalcAfterDialog();

			return TRUE;
		}
		else
		{
			return FALSE;
		}
}

